#!/usr/bin/python
# Copyright (C) 2006-2011, University of Maryland
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/ or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# Author: Nikunj Patel, James Krycka, Paul Kienzle

"""
This module implements the FitControl class which presents a pop-up dialog box
for the user to control fitting options.
"""

# TODO: reset button sets values to factory settings for current optimizer

# ==============================================================================
import wx

import wx.lib.newevent

from .. import options
from .input_list import InputListPanel
from .utilities import phoenix

(FitterChangedEvent, EVT_FITTER_CHANGED) = wx.lib.newevent.NewCommandEvent()


class FitConfig(wx.Frame):
    """
    FitControl lets the user set fitting options from a pop-up dialog box.
    """

    def __init__(
        self,
        parent=None,
        id=wx.ID_ANY,
        title="Fit Options",
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,  # dialog box size will be calculated
        style=wx.DEFAULT_DIALOG_STYLE,
        name="",
        config=None,
        help=None,
        fontsize=None,
    ):
        wx.Frame.__init__(self, parent, id, title, pos, size, style, name)

        self.config = config
        self.help = help

        pairs = [(config.names[id], id) for id in config.active_ids]
        self.active_ids = [id for _, id in sorted(pairs)]

        # Set the font for this window and all child windows (widgets) from the
        # parent window, or from the system defaults if no parent is given.
        # A dialog box does not inherit font info from its parent, so we will
        # explicitly get it from the parent and apply it to the dialog box.
        if parent is not None:
            font = parent.GetFont()
            self.SetFont(font)

        # If the caller specifies a font size, override the default value.
        if fontsize is not None:
            font = self.GetFont()
            font.SetPointSize(fontsize)
            self.SetFont(font)

        # Section 1
        self.panel1 = wx.Panel(self, -1)
        static_box1 = wx.StaticBox(self.panel1, -1, "Fit Algorithms")

        rows = (len(self.active_ids) + 1) // 2

        flexsizer = wx.FlexGridSizer(rows, 2, hgap=20, vgap=10)

        self.fitter_button = {}
        for fitter in self.active_ids:
            button = wx.RadioButton(self.panel1, -1, label=config.names[fitter], name=fitter)
            self.fitter_button[fitter] = button
            self.Bind(wx.EVT_RADIOBUTTON, self.OnRadio, id=button.GetId())
            flexsizer.Add(button, 0, 0)

        fit_hsizer = wx.StaticBoxSizer(static_box1, orient=wx.VERTICAL)
        fit_hsizer.Add(flexsizer, 0, wx.ALL, 5)

        self.panel1.SetSizer(fit_hsizer)

        self.vbox = wx.BoxSizer(wx.VERTICAL)
        self.vbox.Add(self.panel1, 0, wx.ALL, 10)

        # Section 2
        # Create list of all panels for later use in hiding and showing panels.
        self.fitter_panel = {}
        for fitter in self.active_ids:
            items = [
                (options.FIT_FIELDS[field][0], field, config.values[fitter][field], options.FIT_FIELDS[field][1])
                for field, default in config.settings[fitter]
            ]
            # print fitter, items
            panel = ParameterPanel(self, items, config.names[fitter])
            self.fitter_panel[fitter] = panel
            self.vbox.Add(panel, 1, wx.EXPAND | wx.ALL, 10)
            panel.Hide()

        # Make the current panel active
        self.fitter_button[config.selected_id].SetValue(True)
        self.fitter_panel[config.selected_id].Show()

        # Section 3
        # Create the button controls (Reset, Apply) and bind their events.
        # apply_btn = wx.Button(self, wx.ID_APPLY, "Apply")
        # apply_btn.SetToolTip(wx.ToolTip("Accept new options for the optimizer"))
        # apply_btn.SetDefault()
        # reset_btn = wx.Button(self, wx.ID_ANY, "Reset")
        # reset_btn.SetToolTip(wx.ToolTip("Restore default options for the optimizer"))
        accept_btn = wx.Button(self, wx.ID_OK)
        accept_btn.SetToolTip(wx.ToolTip("Accept new options for the optimizer"))
        accept_btn.SetDefault()
        cancel_btn = wx.Button(self, wx.ID_CANCEL)
        cancel_btn.SetToolTip(wx.ToolTip("Restore default options for the optimizer"))
        if help is not None:
            help_btn = wx.Button(self, wx.ID_HELP, "Help")
            # help_btn = wx.Button(self, wx.ID_ANY, 'Help')
            help_btn.SetToolTip(wx.ToolTip("Help on the options for the optimizer"))

        self.Bind(wx.EVT_BUTTON, self.OnAccept, accept_btn)
        self.Bind(wx.EVT_BUTTON, self.OnCancel, cancel_btn)
        if help is not None:
            self.Bind(wx.EVT_BUTTON, self.OnHelp, help_btn)

        self.Bind(wx.EVT_CLOSE, self.OnClose)

        # Create the button sizer that will put the buttons in a row, right
        # justified, and with a fixed amount of space between them.  This
        # emulates the Windows convention for placing a set of buttons at the
        # bottom right of the window.
        btn_sizer = wx.BoxSizer(wx.HORIZONTAL)
        btn_sizer.Add((10, 20), 1)  # stretchable whitespace
        btn_sizer.Add(accept_btn, 0)
        btn_sizer.Add((10, 20), 0)  # non-stretchable whitespace
        btn_sizer.Add(cancel_btn, 0)
        btn_sizer.Add((10, 20), 0)  # non-stretchable whitespace
        if help is not None:
            btn_sizer.Add((10, 20), 0)  # non-stretchable whitespace
            btn_sizer.Add(help_btn, 0)

        # Add the button sizer to the main sizer.
        self.vbox.Add(btn_sizer, 0, wx.EXPAND | wx.ALL, 10)

        # Finalize the sizer and establish the dimensions of the dialog box.
        # The minimum width is explicitly set because the sizer is not able to
        # take into consideration the width of the enclosing frame's title.
        self.SetSizer(self.vbox)
        # self.vbox.SetMinSize((size[0], -1))
        self.vbox.Fit(self)

        self.Centre()

    def OnRadio(self, event):
        button = event.GetEventObject()
        for panel in self.fitter_panel.values():
            panel.Hide()
        self.fitter_panel[button.Name].Show()
        self.vbox.Layout()

    def OnCancel(self, event):
        """
        Restore options for the selected fitter to the default values.
        """
        fitter = self._get_fitter()
        panel = self.fitter_panel[fitter]
        panel.Parameters = dict(self.config.settings[fitter])
        self.Hide()

    def OnAccept(self, event):
        """
        Save the current fitter and options to the fit config.
        """
        fitter = self._get_fitter()
        options = self.fitter_panel[fitter].Parameters
        self.config.selected_id = fitter
        self.config.values[fitter] = options
        self.Hide()

        # Signal a change in fitter
        event = FitterChangedEvent(self.Id, config=self.config)
        wx.PostEvent(self, event)

    def OnHelp(self, event):
        """
        Provide help on the selected fitter.
        """
        if self.help is not None:
            self.help(self._get_fitter())

    def OnClose(self, event):
        """
        Don't close the window, just hide it.
        """
        if event.CanVeto():
            self.Hide()
            event.Veto()
        else:
            event.Skip()

    def _get_fitter(self):
        """
        Returns the currently selected algorithm, or None if no algorithm is
        selected.
        """
        for button in self.fitter_button.values():
            if button.Value:
                return button.Name
        else:
            return None

    def _get_options(self):
        fitter = self._get_fitter()
        options = self.fitter_panel[fitter].Parameters

        return fitter, options


class ParameterPanel(wx.Panel):
    def __init__(self, parent, parameters, fitter_name):
        wx.Panel.__init__(self, parent=parent, id=wx.ID_ANY)

        self.fields = []
        itemlist = []
        sbox = wx.StaticBox(self, wx.ID_ANY, fitter_name + " Fitting Parameters")

        for parameter in parameters:
            label, field, curr_value, datatype = parameter
            if not label.endswith(":"):
                label += ":"
            if hasattr(datatype, "choices"):
                extra = [str(v) for v in datatype.choices]
                mode = "CRE"
            else:
                mode = "RE"
                extra = None
            sub_list = [label, curr_value, datatype, mode, extra]
            itemlist.append(sub_list)
            self.fields.append(field)

        # Set the same minimum height for each panel.  The y-size value should
        # be sufficient to display at least 6 input fields without the need for
        # a scroll bar.  Adjust the size.y value if the maximum number of
        # number of input parameters across fitters changes.
        self.fit_params = InputListPanel(parent=self, itemlist=itemlist, align=True, size=(-1, 220))

        sbox_sizer = wx.StaticBoxSizer(sbox, wx.VERTICAL)
        sbox_sizer.Add(self.fit_params, 1, wx.EXPAND | wx.ALL, 5)
        self.SetSizer(sbox_sizer)
        sbox_sizer.Fit(self)

    @property
    def Parameters(self):
        values = self.fit_params.GetResults()
        return dict(zip(self.fields, values))

    @Parameters.setter
    def Parameters(self, parameters):
        values = [parameters[k] for k in self.fields]
        self.fit_params.update_items_in_panel(values)


_fit_config_frame = None


def show_fit_config(parent, help=None):
    global _fit_config_frame
    if _fit_config_frame is None:
        _fit_config_frame = FitConfig(parent=parent, config=options.FIT_CONFIG, help=help)
    _fit_config_frame.Show()
    _fit_config_frame.Raise()
    return _fit_config_frame


if __name__ == "__main__":
    opts = options.getopts()

    def _help(algo):
        print("asking for help with " + algo)

    app = wx.App()
    top = wx.Frame(None)
    text = wx.TextCtrl(top, wx.ID_ANY, "some text")
    button = wx.Button(top, wx.ID_ANY, "Options...")
    button.Bind(wx.EVT_BUTTON, lambda ev: show_fit_config(top, help=_help))

    sizer = wx.BoxSizer(wx.VERTICAL)
    sizer.Add(text)
    sizer.Add(button)
    sizer.Fit(top)
    top.SetSizer(sizer)
    top.Show()
    app.MainLoop()



"""
Sample models and functions prepared for use in mystic


Functions
=========

    corana 1d,2d,3d,4d     -- Corana's function


Models
======

    dense_circle     -- 2d array representation of a circle
    sparse_circle
    minimal_circle
    decay            -- Bevington & Robinson's model of dual exponential decay

"""

# base classes
# from .model import Fitness, Function

# models
# from .decay import decay
# from .circle import dense_circle, sparse_circle, minimal_circle

# functions
# from .corana import corana1d, corana2d, corana3d, corana4d

# end of file



# This program is in the public domain
# Author: Paul Kienzle
"""
Parameter bounds and prior probabilities.

Parameter bounds encompass several features of our optimizers.

First and most trivially they allow for bounded constraints on
parameter values.

Secondly, for parameter values known to follow some distribution,
the bounds encodes a penalty function as the value strays from
its nominal value.  Using a negative log likelihood cost function
on the fit, then this value naturally contributes to the overall
likelihood measure.

Predefined bounds are::

    Unbounded
        range (-inf, inf)
    BoundedBelow
        range (base, inf)
    BoundedAbove
        range (-inf, base)
    Bounded
        range (low, high)
    Normal
        range (-inf, inf) with gaussian probability
    BoundedNormal
        range (low, high) with gaussian probability within
    SoftBounded
        range (low, high) with gaussian probability outside

New bounds can be defined following the abstract base class
interface defined in :class:`Bounds`, or using Distribution(rv)
where rv is a scipy.stats continuous distribution.

For generating bounds given a value, we provide a few helper
functions::

    v +/- d:  pm(x,dx) or pm(x,-dm,+dp) or pm(x,+dp,-dm)
        return (x-dm,x+dm) limited to 2 significant digits
    v +/- p%: pmp(x,p) or pmp(x,-pm,+pp) or pmp(x,+pp,-pm)
        return (x-pm*x/100, x+pp*x/100) limited to 2 sig. digits
    pm_raw(x,dx) or raw_pm(x,-dm,+dp) or raw_pm(x,+dp,-dm)
        return (x-dm,x+dm)
    pmp_raw(x,p) or raw_pmp(x,-pm,+pp) or raw_pmp(x,+pp,-pm)
        return (x-pm*x/100, x+pp*x/100)
    nice_range(lo,hi)
        return (lo,hi) limited to 2 significant digits
"""

__all__ = [
    "pm",
    "pmp",
    "pm_raw",
    "pmp_raw",
    "nice_range",
    "init_bounds",
    "DistProtocol",
    "Bounds",
    "Unbounded",
    "Bounded",
    "BoundedAbove",
    "BoundedBelow",
    "Distribution",
    "Normal",
    "BoundedNormal",
    "SoftBounded",
]

from dataclasses import dataclass, field
import math
from math import log, log10, sqrt, pi, ceil, floor

from numpy import inf, isinf, isfinite, clip
import numpy.random as RNG

try:
    from scipy.stats import norm as normal_distribution
except ImportError:
    # Normal distribution is an optional dependency.  Leave it as a runtime
    # failure if it doesn't exist.
    pass

from typing import Optional, Any, Dict, Union, Literal, Tuple, Protocol

LimitValue = Union[float, Literal["-inf", "inf"]]
LimitsType = Tuple[Union[float, Literal["-inf"]], Union[float, Literal["inf"]]]

# TODO: should we use this in the bounds limits?
# @dataclass(init=False)
# class ExtendedFloat(float):
#     __root__: Union[float, Literal["inf", "-inf"]]
#     def __new__(cls, *args, **kw):
#         return super().__new__(cls, *args)
#     def __init__(self, __root__=None):
#         pass
#     def __repr__(self):
#         return float.__repr__(self)


def pm(v, plus, minus=None, limits: Optional[LimitsType] = None):
    """
    Return the tuple (~v-dv,~v+dv), where ~expr is a 'nice' number near to
    to the value of expr.  For example::

        >>> r = pm(0.78421, 0.0023145)
        >>> print("%g - %g"%r)
        0.7818 - 0.7866

    If called as pm(value, +dp, -dm) or pm(value, -dm, +dp),
    return (~v-dm, ~v+dp).
    """
    return nice_range(limited_range(pm_raw(v, plus, minus), limits=limits))


def pmp(v, plus, minus=None, limits=None):
    """
    Return the tuple (~v-%v,~v+%v), where ~expr is a 'nice' number near to
    the value of expr.  For example::

        >>> r = pmp(0.78421, 10)
        >>> print("%g - %g"%r)
        0.7 - 0.87
        >>> r = pmp(0.78421, 0.1)
        >>> print("%g - %g"%r)
        0.7834 - 0.785

    If called as pmp(value, +pp, -pm) or pmp(value, -pm, +pp),
    return (~v-pm%v, ~v+pp%v).
    """
    return nice_range(limited_range(pmp_raw(v, plus, minus), limits=limits))


# Generate ranges using x +/- dx or x +/- p%*x


def pm_raw(v, plus, minus=None):
    """
    Return the tuple [v-dv,v+dv].

    If called as pm_raw(value, +dp, -dm) or pm_raw(value, -dm, +dp),
    return (v-dm, v+dp).
    """
    if minus is None:
        minus = -plus
    if plus < minus:
        plus, minus = minus, plus
    return v + minus, v + plus


def pmp_raw(v, plus, minus=None):
    """
    Return the tuple [v-%v,v+%v]

    If called as pmp_raw(value, +pp, -pm) or pmp_raw(value, -pm, +pp),
    return (v-pm%v, v+pp%v).
    """
    if minus is None:
        minus = -plus
    if plus < minus:
        plus, minus = minus, plus
    b1, b2 = v * (1 + 0.01 * minus), v * (1 + 0.01 * plus)

    return (b1, b2) if v > 0 else (b2, b1)


def limited_range(bounds, limits=None):
    """
    Given a range and limits, fix the endpoints to lie within the range
    """
    if limits is not None:
        return clip(bounds[0], *limits), clip(bounds[1], *limits)
    return bounds


def nice_range(bounds):
    """
    Given a range, return an enclosing range accurate to two digits.
    """
    step = bounds[1] - bounds[0]
    if step > 0:
        d = 10 ** (floor(log10(step)) - 1)
        return floor(bounds[0] / d) * d, ceil(bounds[1] / d) * d
    else:
        return bounds


def init_bounds(v) -> "Bounds":
    """
    Returns a bounds object of the appropriate type given the arguments.

    This is a helper factory to simplify the user interface to parameter
    objects.
    """
    # if it is none, then it is unbounded
    if v is None:
        return Unbounded()

    # if it isn't a tuple, assume it is a bounds type.
    try:
        lo, hi = v
    except TypeError:
        return v

    # if it is a tuple, then determine what kind of bounds we have
    if lo is None:
        lo = -inf
    if hi is None:
        hi = inf
    # TODO: consider issuing a warning instead of correcting reversed bounds
    if lo >= hi:
        lo, hi = hi, lo
    if isinf(lo) and isinf(hi):
        return Unbounded()
    elif isinf(lo):
        return BoundedAbove(hi)
    elif isinf(hi):
        return BoundedBelow(lo)
    else:
        return Bounded(lo, hi)


class Bounds:
    """
    Bounds abstract base class.

    A range is used for several purposes.  One is that it transforms parameters
    between unbounded and bounded forms depending on the needs of the optimizer.

    Another is that it generates random values in the range for stochastic
    optimizers, and for initialization.

    A third is that it returns the likelihood of seeing that particular value
    for optimizers which use soft constraints.  Assuming the cost function that
    is being optimized is also a probability, then this is an easy way to
    incorporate information from other sorts of measurements into the model.
    """

    # TODO: need derivatives wrt bounds transforms

    @property
    def limits(self):
        return (-inf, inf)

    def get01(self, x):
        """
        Convert value into [0,1] for optimizers which are bounds constrained.

        This can also be used as a scale bar to show approximately how close to
        the end of the range the value is.
        """

    def put01(self, v):
        """
        Convert [0,1] into value for optimizers which are bounds constrained.
        """

    def getfull(self, x):
        """
        Convert value into (-inf,inf) for optimizers which are unconstrained.
        """

    def putfull(self, v):
        """
        Convert (-inf,inf) into value for optimizers which are unconstrained.
        """

    def random(self, n=1, target=1.0):
        """
        Return a randomly generated valid value.

        *target* gives some scale independence to the random number
        generator, allowing the initial value of the parameter to influence
        the randomly generated value.  Otherwise fits without bounds have
        too large a space to search through.
        """

    def nllf(self, value):
        """
        Return the negative log likelihood of seeing this value, with
        likelihood scaled so that the maximum probability is one.

        For uniform bounds, this either returns zero or inf.  For bounds
        based on a probability distribution, this returns values between
        zero and inf.  The scaling is necessary so that indefinite and
        semi-definite ranges return a sensible value.  The scaling does
        not affect the likelihood maximization process, though the resulting
        likelihood is not easily interpreted.
        """

    def residual(self, value):
        """
        Return the parameter 'residual' in a way that is consistent with
        residuals in the normal distribution.  The primary purpose is to
        graphically display exceptional values in a way that is familiar
        to the user.  For fitting, the scaled likelihood should be used.

        To do this, we will match the cumulative density function value
        with that for N(0,1) and find the corresponding percent point
        function from the N(0,1) distribution.  In this way, for example,
        a value to the right of 2.275% of the distribution would correspond
        to a residual of -2, or 2 standard deviations below the mean.

        For uniform distributions, with all values equally probable, we
        use a value of +/-4 for values outside the range, and 0 for values
        inside the range.
        """

    def start_value(self):
        """
        Return a default starting value if none given.
        """
        return self.put01(0.5)

    def __contains__(self, v):
        return self.limits[0] <= v <= self.limits[1]

    def __str__(self):
        limits = tuple(num_format(v) for v in self.limits)
        return "(%s,%s)" % limits

    def satisfied(self, v) -> bool:
        lo, hi = self.limits
        return v >= lo and v <= hi

    def penalty(self, v) -> float:
        """
        return a (differentiable) nonzero value when outside the bounds
        """
        lo, hi = self.limits
        dlo = 0.0 if v >= lo else abs(v - lo)
        dhi = 0.0 if v <= hi else abs(v - hi)
        return dlo + dhi

    def to_dict(self):
        return dict(
            type=type(self).__name__,
            limits=self.limits,
        )


# CRUFT: python 2.5 doesn't format indefinite numbers properly on windows


def num_format(v):
    """
    Number formating which supports inf/nan on windows.
    """
    if isfinite(v):
        return "%g" % v
    elif isinf(v):
        return "inf" if v > 0 else "-inf"
    else:
        return "NaN"


@dataclass(init=False)
class Unbounded(Bounds):
    """
    Unbounded parameter.

    The random initial condition is assumed to be between 0 and 1

    The probability is uniformly 1/inf everywhere, which means the negative
    log likelihood of P is inf everywhere.  A value inf will interfere
    with optimization routines, and so we instead choose P == 1 everywhere.
    """

    type = "Unbounded"

    def __init__(self, *args, **kw):
        pass

    def random(self, n=1, target=1.0):
        scale = target + (target == 0.0)
        return RNG.randn(n) * scale

    def nllf(self, value):
        return 0

    def residual(self, value):
        return 0

    def get01(self, x):
        return _get01_inf(x)

    def put01(self, v):
        return _put01_inf(v)

    def getfull(self, x):
        return x

    def putfull(self, v):
        return v


@dataclass(init=True)
class BoundedBelow(Bounds):
    """
    Semidefinite range bounded below.

    The random initial condition is assumed to be within 1 of the maximum.

    [base,inf] <-> (-inf,inf) is direct above base+1, -1/(x-base) below
    [base,inf] <-> [0,1] uses logarithmic compression.

    Logarithmic compression works by converting sign*m*2^e+base to
    sign*(e+1023+m), yielding a value in [0,2048]. This can then be
    converted to a value in [0,1].

    Note that the likelihood function is problematic: the true probability
    of seeing any particular value in the range is infinitesimal, and that
    is indistinguishable from values outside the range.  Instead we say
    that P = 1 in range, and 0 outside.
    """

    base: float
    type = "BoundedBelow"

    @property
    def limits(self):
        return (self.base, inf)

    def start_value(self):
        return self.base + 1

    def random(self, n=1, target: float = 1.0):
        target = max(abs(target), abs(self.base))
        scale = target + float(target == 0.0)
        return self.base + abs(RNG.randn(n) * scale)

    def nllf(self, value):
        return 0 if value >= self.base else inf

    def residual(self, value):
        return 0 if value >= self.base else -4

    def get01(self, x):
        m, e = math.frexp(x - self.base)
        if m >= 0 and e <= _E_MAX:
            v = (e + m) / (2.0 * _E_MAX)
            return v
        else:
            return 0 if m < 0 else 1

    def put01(self, v):
        v = v * 2 * _E_MAX
        e = int(v)
        m = v - e
        x = math.ldexp(m, e) + self.base
        return x

    def getfull(self, x):
        v = x - self.base
        return v if v >= 1 else 2 - 1.0 / v

    def putfull(self, v):
        x = v if v >= 1 else 1.0 / (2 - v)
        return x + self.base


@dataclass(init=True)
class BoundedAbove(Bounds):
    """
    Semidefinite range bounded above.

    [-inf,base] <-> [0,1] uses logarithmic compression
    [-inf,base] <-> (-inf,inf) is direct below base-1, 1/(base-x) above

    Logarithmic compression works by converting sign*m*2^e+base to
    sign*(e+1023+m), yielding a value in [0,2048].  This can then be
    converted to a value in [0,1].

    Note that the likelihood function is problematic: the true probability
    of seeing any particular value in the range is infinitesimal, and that
    is indistinguishable from values outside the range.  Instead we say
    that P = 1 in range, and 0 outside.
    """

    base: float

    @property
    def limits(self):
        return (-inf, self.base)

    def start_value(self):
        return self.base - 1

    def random(self, n=1, target: float = 1.0):
        target = max(abs(self.base), abs(target))
        scale = target + float(target == 0.0)
        return self.base - abs(RNG.randn(n) * scale)

    def nllf(self, value):
        return 0 if value <= self.base else inf

    def residual(self, value):
        return 0 if value <= self.base else 4

    def get01(self, x):
        m, e = math.frexp(self.base - x)
        if m >= 0 and e <= _E_MAX:
            v = (e + m) / (2.0 * _E_MAX)
            return 1 - v
        else:
            return 1 if m < 0 else 0

    def put01(self, v):
        v = (1 - v) * 2 * _E_MAX
        e = int(v)
        m = v - e
        x = -(math.ldexp(m, e) - self.base)
        return x

    def getfull(self, x):
        v = x - self.base
        return v if v <= -1 else -2 - 1.0 / v

    def putfull(self, v):
        x = v if v <= -1 else -1.0 / (v + 2)
        return x + self.base


@dataclass(init=True)
class Bounded(Bounds):
    """
    Bounded range.

    [lo,hi] <-> [0,1] scale is simple linear
    [lo,hi] <-> (-inf,inf) scale uses exponential expansion

    While technically the probability of seeing any value within the
    range is 1/range, for consistency with the semi-infinite ranges
    and for a more natural mapping between nllf and chisq, we instead
    set the probability to 0.  This choice will not affect the fits.
    """

    lo: float = field(metadata={"description": "lower end of bounds"})
    hi: float = field(metadata={"description": "upper end of bounds"})

    # @classmethod
    # def from_dict(cls, limits=None):
    #     lo, hi = limits
    #     return cls(lo, hi)

    # def __init__(self, lo, hi):
    #     self.lo = lo
    #     self.hi = hi
    #     self._nllf_scale = log(hi - lo)

    @property
    def limits(self):
        return (self.lo, self.hi)

    def random(self, n=1, target=1.0):
        # print("= uniform",lo,hi)
        return RNG.uniform(self.lo, self.hi, size=n)

    def nllf(self, value):
        return 0 if self.lo <= value <= self.hi else inf
        # return self._nllf_scale if lo<=value<=hi else inf

    def residual(self, value):
        return -4 if self.lo > value else (4 if self.hi < value else 0)

    def get01(self, x):
        lo, hi = self.limits
        return float(x - lo) / (hi - lo) if hi - lo > 0 else 0

    def put01(self, v):
        lo, hi = self.limits
        return (hi - lo) * v + lo

    def getfull(self, x):
        return _put01_inf(self.get01(x))

    def putfull(self, v):
        return self.put01(_get01_inf(v))


class DistProtocol(Protocol):
    """
    Protocol for a distribution object, implementing the scipy.stats interface.
    (also including args, kwds and name)
    """

    name: str
    args: Tuple[float, ...]
    kwds: Dict[str, Any]

    def rvs(self, n: int) -> float: ...
    def nnlf(self, value: float) -> float: ...
    def cdf(self, value: float) -> float: ...
    def ppf(self, value: float) -> float: ...
    def pdf(self, value: float) -> float: ...


class Distribution(Bounds):
    """
    Parameter is pulled from a distribution.

    *dist* must implement the distribution interface from scipy.stats,
    described in the DistProtocol class.
    """

    dist: DistProtocol = None

    def __init__(self, dist):
        object.__setattr__(self, "dist", dist)

    def random(self, n=1, target=1.0):
        return self.dist.rvs(n)

    def nllf(self, value):
        return -log(self.dist.pdf(value))

    def residual(self, value):
        return normal_distribution.ppf(self.dist.cdf(value))

    def get01(self, x):
        return self.dist.cdf(x)

    def put01(self, v):
        return self.dist.ppf(v)

    def getfull(self, x):
        return x

    def putfull(self, v):
        return v

    def __getstate__(self):
        # WARNING: does not preserve and restore seed
        return self.dist.__class__, self.dist.args, self.dist.kwds

    def __setstate__(self, state):
        cls, args, kwds = state
        self.dist = cls(*args, **kwds)

    def __str__(self):
        return "%s(%s)" % (self.dist.dist.name, ",".join(str(s) for s in self.dist.args))

    def to_dict(self):
        return dict(
            type=type(self).__name__,
            limits=self.limits,
            # TODO: how to handle arbitrary distribution function in save/load?
            dist=type(self.dist).__name__,
        )


@dataclass(frozen=True)
class Normal(Distribution):
    """
    Parameter is pulled from a normal distribution.

    If you have measured a parameter value with some uncertainty (e.g., the
    film thickness is 35+/-5 according to TEM), then you can use this
    measurement to restrict the values given to the search, and to penalize
    choices of this fitting parameter which are different from this value.

    *mean* is the expected value of the parameter and *std* is the 1-sigma
    standard deviation.

    class is 'frozen' because a new object should be created if
    `mean` or `std` are changed.
    """

    mean: float = 0.0
    std: float = 1.0
    _nllf_scale: float = field(init=False)

    def __post_init__(self):
        Distribution.__init__(self, normal_distribution(self.mean, self.std))
        object.__setattr__(self, "_nllf_scale", log(2 * pi * self.std**2) / 2)

    def nllf(self, value):
        # P(v) = exp(-0.5*(v-mean)**2/std**2)/sqrt(2*pi*std**2)
        # -log(P(v)) = -(-0.5*(v-mean)**2/std**2 - log( (2*pi*std**2) ** 0.5))
        #            = 0.5*(v-mean)**2/std**2 + log(2*pi*std**2)/2
        mean, std = self.dist.args
        return 0.5 * ((value - mean) / std) ** 2 + self._nllf_scale

    def residual(self, value):
        mean, std = self.dist.args
        return (value - mean) / std

    def __getstate__(self):
        return self.dist.args  # args is mean,std

    def __setstate__(self, state):
        mean, std = state
        self.__init__(mean=mean, std=std)


@dataclass(init=False, frozen=True)
class BoundedNormal(Bounds):
    """
    truncated normal bounds
    """

    mean: float = 0.0
    std: float = 1.0
    lo: Union[float, Literal["-inf"]]
    hi: Union[float, Literal["inf"]]

    _left: float = field(init=False)
    _delta: float = field(init=False)
    _nllf_scale: float = field(init=False)

    def __init__(self, mean: float = 0, std: float = 1, limits=(-inf, inf), hi="inf", lo="-inf"):
        if limits is not None:
            # for backward compatibility:
            lo, hi = limits
        limits = (-inf if lo is None else float(lo), inf if hi is None else float(hi))
        object.__setattr__(self, "lo", limits[0])
        object.__setattr__(self, "hi", limits[1])
        object.__setattr__(self, "mean", mean)
        object.__setattr__(self, "std", std)

        object.__setattr__(self, "_left", normal_distribution.cdf((limits[0] - mean) / std))
        object.__setattr__(self, "_delta", normal_distribution.cdf((limits[1] - mean) / std) - self._left)
        object.__setattr__(self, "_nllf_scale", log(2 * pi * std**2) / 2 + log(self._delta))

    @property
    def limits(self):
        return (self.lo, self.hi)

    def get01(self, x):
        """
        Convert value into [0,1] for optimizers which are bounds constrained.

        This can also be used as a scale bar to show approximately how close to
        the end of the range the value is.
        """
        v = (normal_distribution.cdf((x - self.mean) / self.std) - self._left) / self._delta
        return clip(v, 0, 1)

    def put01(self, v):
        """
        Convert [0,1] into value for optimizers which are bounds constrained.
        """
        x = v * self._delta + self._left
        return normal_distribution.ppf(x) * self.std + self.mean

    def getfull(self, x):
        """
        Convert value into (-inf,inf) for optimizers which are unconstrained.
        """
        raise NotImplementedError

    def putfull(self, v):
        """
        Convert (-inf,inf) into value for optimizers which are unconstrained.
        """
        raise NotImplementedError

    def random(self, n=1, target=1.0):
        """
        Return a randomly generated valid value, or an array of values
        """
        return self.get01(RNG.rand(n))

    def nllf(self, value):
        """
        Return the negative log likelihood of seeing this value, with
        likelihood scaled so that the maximum probability is one.
        """
        if value in self:
            return 0.5 * ((value - self.mean) / self.std) ** 2 + self._nllf_scale
        else:
            return inf

    def residual(self, value):
        """
        Return the parameter 'residual' in a way that is consistent with
        residuals in the normal distribution.  The primary purpose is to
        graphically display exceptional values in a way that is familiar
        to the user.  For fitting, the scaled likelihood should be used.

        For the truncated normal distribution, we can just use the normal
        residuals.
        """
        return (value - self.mean) / self.std

    def start_value(self):
        """
        Return a default starting value if none given.
        """
        return self.put01(0.5)

    def __contains__(self, v):
        return self.limits[0] <= v <= self.limits[1]

    def __str__(self):
        vals = (
            self.limits[0],
            self.limits[1],
            self.mean,
            self.std,
        )
        return "(%s,%s), norm(%s,%s)" % tuple(num_format(v) for v in vals)


@dataclass(init=False, frozen=True)
class SoftBounded(Bounds):
    """
    Parameter is pulled from a stretched normal distribution.

    This is like a rectangular distribution, but with gaussian tails.

    The intent of this distribution is for soft constraints on the values.
    As such, the random generator will return values like the rectangular
    distribution, but the likelihood will return finite values based on
    the distance from the from the bounds rather than returning infinity.

    Note that for bounds constrained optimizers which force the value
    into the range [0,1] for each parameter we don't need to use soft
    constraints, and this acts just like the rectangular distribution.
    """

    lo: float = 0.0
    hi: float = 1.0
    std: float = 1.0

    def __init__(self, lo, hi, std=1.0):
        self.lo, self.hi, self.std = lo, hi, std
        self._nllf_scale = log(hi - lo + sqrt(2 * pi * std))

    @property
    def limits(self):
        return (self.lo, self.hi)

    def random(self, n=1, target=1.0):
        return RNG.uniform(self.lo, self.hi, size=n)

    def nllf(self, value):
        # To turn f(x) = 1 if x in [lo,hi] else G(tail)
        # into a probability p, we need to normalize by \int{f(x)dx},
        # which is just hi-lo + sqrt(2*pi*std**2).
        if value < self.lo:
            z = self.lo - value
        elif value > self.hi:
            z = value - self.hi
        else:
            z = 0
        return (z / self.std) ** 2 / 2 + self._nllf_scale

    def residual(self, value):
        if value < self.lo:
            z = self.lo - value
        elif value > self.hi:
            z = value - self.hi
        else:
            z = 0
        return z / self.std

    def get01(self, x):
        v = float(x - self.lo) / (self.hi - self.lo)
        return v if 0 <= v <= 1 else (0 if v < 0 else 1)

    def put01(self, v):
        return v * (self.hi - self.lo) + self.lo

    def getfull(self, x):
        return x

    def putfull(self, v):
        return v

    def __str__(self):
        return "box_norm(%g,%g,sigma=%g)" % (self.lo, self.hi, self.std)


_E_MIN = -1023
_E_MAX = 1024


def _get01_inf(x):
    """
    Convert a floating point number to a value in [0,1].

    The value sign*m*2^e to sign*(e+1023+m), yielding a value in [-2048,2048].
    This can then be converted to a value in [0,1].

    Sort order is preserved.  At least 14 bits of precision are lost from
    the 53 bit mantissa.
    """
    # Arctan alternative
    # Arctan is approximately linear in (-0.5, 0.5), but the
    # transform is only useful up to (-10**15,10**15).
    # return atan(x)/pi + 0.5
    m, e = math.frexp(x)
    s = math.copysign(1.0, m)
    v = (e - _E_MIN + m * s) * s
    v = v / (4 * _E_MAX) + 0.5
    v = 0 if _E_MIN > e else (1 if _E_MAX < e else v)
    return v


def _put01_inf(v):
    """
    Convert a value in [0,1] to a full floating point number.

    Sort order is preserved.  Reverses :func:`_get01_inf`, but with fewer
    bits of precision.
    """
    # Arctan alternative
    # return tan(pi*(v-0.5))

    v = (v - 0.5) * 4 * _E_MAX
    s = math.copysign(1.0, v)
    v *= s
    e = int(v)
    m = v - e
    x = math.ldexp(s * m, e + _E_MIN)
    # print "< x,e,m,s,v",x,e+_e_min,s*m,s,v
    return x


BoundsType = Union[Unbounded, Bounded, BoundedAbove, BoundedBelow, BoundedNormal, SoftBounded, Normal]


def test_normal():
    """
    Test the normal distribution
    """
    epsilon = 1e-10
    n = Normal(mean=0.5, std=1.0)
    assert abs(n.nllf(0.5) - 0.9189385332046727) < epsilon
    assert abs(n.nllf(1.0) - n.nllf(0.0)) < epsilon
    assert abs(n.residual(0.5) - 0.0) < epsilon
    assert abs(n.residual(1.0) - 0.5) < epsilon



from functools import lru_cache
import itertools
from types import GeneratorType
from typing import (
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    List,
    Literal,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    Union,
    Tuple,
    TypedDict,
    cast,
)
from datetime import datetime
import numbers
import warnings
from threading import Event
import numpy as np
import asyncio
from pathlib import Path, PurePath
import json
from copy import deepcopy
import os
import uuid
import traceback

from bumps.fitters import (
    DreamFit,
    LevenbergMarquardtFit,
    SimplexFit,
    DEFit,
    MPFit,
    BFGSFit,
    FitDriver,
    fit,
    nllf_scale,
    format_uncertainty,
)
from bumps.mapper import MPMapper
from bumps.parameter import Parameter, Constant, Variable, unique
import bumps.cli
import bumps.fitproblem
import bumps.dream.views, bumps.dream.varplot, bumps.dream.stats, bumps.dream.state
import bumps.errplot

from .state_hdf5_backed import (
    UNDEFINED,
    UNDEFINED_TYPE,
    State,
    serialize_problem,
    deserialize_problem,
    SERIALIZER_EXTENSIONS,
)
from .fit_thread import FitThread, EVT_FIT_COMPLETE, EVT_FIT_PROGRESS
from .varplot import plot_vars
from .traceplot import plot_trace
from .logger import logger, console_handler
from .custom_plot import process_custom_plot, CustomWebviewPlot

REGISTRY: Dict[str, Callable] = {}
MODEL_EXT = ".json"
TRACE_MEMORY = False

FITTERS = (DreamFit, LevenbergMarquardtFit, SimplexFit, DEFit, MPFit, BFGSFit)
FITTERS_BY_ID = dict([(fitter.id, fitter) for fitter in FITTERS])
FITTER_DEFAULTS = {}
for fitter in FITTERS:
    FITTER_DEFAULTS[fitter.id] = {
        "name": fitter.name,
        "settings": dict(fitter.settings),
    }

state = State()


def register(fn: Callable):
    REGISTRY[fn.__name__] = fn
    return fn


class Emitter(Protocol):
    def __call__(
        self,
        event: str,
        data: Optional[Any] = None,
        to: Optional[str] = None,
        room: Optional[str] = None,
        skip_sid: Optional[str] = None,
        namespace: Optional[str] = None,
        callback: Optional[Callable] = None,
        ignore_queue: bool = False,
    ) -> Awaitable: ...


EMITTERS: Dict[str, Emitter] = {}


async def emit(
    event: str,
    data: Optional[Any] = None,
    to: Optional[str] = None,
    room: Optional[str] = None,
    skip_sid: Optional[str] = None,
    namespace: Optional[str] = None,
    callback: Optional[Callable] = None,
    ignore_queue: bool = False,
):
    results = {}
    for emitter_name, emitter_fn in EMITTERS.items():
        results[emitter_name] = await emitter_fn(
            event,
            data=data,
            to=to,
            room=room,
            skip_sid=skip_sid,
            namespace=namespace,
            callback=callback,
            ignore_queue=ignore_queue,
        )
    return results


TopicNameType = Literal[
    "log",  # log messages
]


@register
async def load_problem_file(pathlist: List[str], filename: str, autosave_previous: bool = True):
    path = Path(*pathlist, filename)
    logger.info(f"Loading model: {path}")
    await log(f"Loading model: {path}")
    if filename.endswith(".json"):
        with open(path, "rt") as input_file:
            serialized = input_file.read()
        problem = deserialize_problem(serialized, method="dataclass")
    else:
        from bumps.cli import load_model

        problem = load_model(str(path))
    assert isinstance(problem, bumps.fitproblem.FitProblem)
    # problem_state = ProblemState(problem, pathlist, filename)
    try:
        _serialized_problem = serialize_problem(problem, method="dataclass")
        state.problem.serializer = "dataclass"
    except Exception as exc:
        logger.info(f"Could not serialize problem as JSON (dataclass): {exc}, switching to dill")
        state.problem.serializer = "dill"
    if (
        state.shared.autosave_history
        and autosave_previous
        and state.problem is not None
        and state.problem.fitProblem is not None
    ):
        await save_to_history("autosaved before loading new model")
    state.shared.model_file = dict(filename=filename, pathlist=pathlist)
    state.shared.model_loaded = now_string()
    await set_problem(problem, Path(*pathlist), filename)


@register
async def set_serialized_problem(serialized, new_model: bool = False, name: Optional[str] = None):
    fitProblem = deserialize_problem(serialized, method="dataclass")
    state.problem.serializer = "dataclass"
    await set_problem(fitProblem, new_model=new_model, name=name)


async def set_problem(
    problem: bumps.fitproblem.FitProblem,
    path: Optional[Path] = None,
    filename: str = "",
    new_model: bool = True,
    name: Optional[str] = None,
):
    if state.problem is None or state.problem.fitProblem is None:
        update = False
    state.problem.fitProblem = problem
    name = name if name is not None else problem.name
    if name is None:
        name = filename
    state.shared.updated_model = now_string()
    state.shared.updated_parameters = now_string()

    if new_model:
        pathlist = list(path.parts) if path is not None else []
        path_string = "(no path)" if path is None else str(path / filename)
        await log(f"Model loaded: {path_string}")
        state.shared.model_file = dict(filename=filename, pathlist=pathlist)
        state.shared.model_loaded = now_string()
        if state.shared.autosave_history and state.problem is not None and state.problem.fitProblem is not None:
            await save_to_history(f"Loaded model: {name}", keep=True)
        await add_notification(content=path_string, title="Model loaded", timeout=2000)
    state.autosave()


@register
async def get_history():
    return state.get_history()


@register
async def remove_history_item(timestamp: str):
    state.remove_history_item(timestamp)
    state.shared.updated_history = now_string()


@register
async def save_to_history(
    label: str,
    include_population: bool = False,
    include_uncertainty: bool = False,
    keep: bool = False,
):
    state.save_to_history(
        label,
        include_population=include_population,
        include_uncertainty=include_uncertainty,
        keep=keep,
    )


@register
async def reload_history_item(timestamp: str):
    state.reload_history_item(timestamp)


@register
async def set_keep_history(timestamp: str, keep: bool):
    state.history.set_keep(timestamp, keep)
    state.shared.updated_history = now_string()


@register
async def update_history_label(timestamp: str, label: str):
    state.history.update_label(timestamp, label)
    state.shared.updated_history = now_string()


@register
async def save_problem_file(
    pathlist: Optional[List[str]] = None,
    filename: Optional[str] = None,
    overwrite: bool = False,
):
    problem_state = state.problem
    if problem_state is None:
        logger.warning("Save failed: no problem loaded.")
        return
    if pathlist is None:
        pathlist = problem_state.pathlist
    if filename is None:
        filename = problem_state.filename

    if pathlist is None or filename is None:
        logger.warning("no filename and path provided to save")
        return {"filename": "", "check_overwrite": False}

    path = Path(*pathlist)
    serializer = state.problem.serializer
    extension = SERIALIZER_EXTENSIONS[serializer]
    save_filename = f"{Path(filename).stem}.{extension}"
    if not overwrite and Path.exists(path / save_filename):
        # confirmation needed:
        return {"filename": save_filename, "check_overwrite": True}

    serialized = serialize_problem(problem_state.fitProblem, method=serializer)
    with open(Path(path, save_filename), "wb") as output_file:
        output_file.write(serialized)

    await log(f"Saved: {save_filename} at path: {path}")
    return {"filename": save_filename, "check_overwrite": False}


@register
async def save_session():
    state.save()


@register
async def save_session_copy(pathlist: List[str], filename: str):
    path = Path(*pathlist)
    state.write_session_file(str(path / filename))


@register
async def load_session(pathlist: List[str], filename: str, read_only: bool = False):
    path = Path(*pathlist)
    state.setup_backing(filename, pathlist, read_only=read_only)
    state.shared.updated_model = now_string()
    state.shared.updated_parameters = now_string()


@register
async def set_session_output_file(pathlist: Optional[List[str]] = None, filename: Optional[str] = None):
    if filename is None or pathlist is None:
        await state.shared.set("session_output_file", None)
    else:
        await state.shared.set("session_output_file", dict(filename=filename, pathlist=pathlist))


@register
async def get_serializer():
    output = {"serializer": "", "extension": ""}
    problem_state = state.problem
    if problem_state is not None:
        serializer = problem_state.serializer
        output["serializer"] = serializer
        output["extension"] = SERIALIZER_EXTENSIONS[serializer]
    return output


@register
async def export_results(export_path: Union[str, List[str]] = ""):
    from concurrent.futures import ThreadPoolExecutor

    problem_state = state.problem
    if problem_state is None:
        logger.warning("Save failed: no problem loaded.")
        return

    problem = deepcopy(problem_state.fitProblem)
    # TODO: if making a temporary copy of the uncertainty state is going to cause memory
    # issues, we could try to copy and then fall back to just using the live object,
    # or we could just always use the live object, which is unlikely to be changed before
    # the export completes, anyway.
    uncertainty_state = deepcopy(state.fitting.uncertainty_state)

    if not isinstance(export_path, list):
        export_path = [export_path]
    path = Path(*export_path)
    notification_id = await add_notification(content=f"<span>{str(path)}</span>", title="Export started", timeout=None)
    try:
        await asyncio.to_thread(_export_results, path, problem, uncertainty_state)
    finally:
        await emit("cancel_notification", notification_id)


def _export_results(
    path: Path,
    problem: bumps.fitproblem.FitProblem,
    uncertainty_state: Optional[bumps.dream.state.MCMCDraw],
):
    from bumps.util import redirect_console

    basename = problem.name if problem.name is not None else "problem"
    # Storage directory
    path.mkdir(parents=True, exist_ok=True)
    output_pathstr = str(path / basename)

    # Ask model to save its information
    problem.save(output_pathstr)

    # Save a snapshot of the model that can (hopefully) be reloaded
    serializer = state.problem.serializer
    extension = SERIALIZER_EXTENSIONS[serializer]
    save_filename = f"{output_pathstr}.{extension}"
    try:
        serialized = serialize_problem(problem, serializer)
        with open(save_filename, "wb") as output_file:
            output_file.write(serialized)
    except Exception as exc:
        logger.error(f"Error exporting model: {exc}")

    # Save the current state of the parameters
    with redirect_console(str(path / f"{basename}.out")):
        problem.show()

    pardata = "".join("%s %.15g\n" % (name, value) for name, value in zip(problem.labels(), problem.getp()))

    open(path / f"{basename}.par", "wt").write(pardata)

    # Produce model plots
    problem.plot(figfile=output_pathstr)

    # Produce uncertainty plots
    if uncertainty_state is not None:
        with redirect_console(str(path / f"{basename}.err")):
            uncertainty_state.show(figfile=output_pathstr)
        uncertainty_state.save(output_pathstr)


@register
async def save_parameters(pathlist: List[str], filename: str, overwrite: bool = False):
    problem_state = state.problem
    if problem_state is None:
        await log("Error: Can't save parameters if no problem loaded")
        return
    problem = problem_state.fitProblem
    path = Path(*pathlist)
    if not overwrite and Path.exists(path / filename):
        # confirmation needed:
        return {"filename": filename, "check_overwrite": True}

    pardata = "".join("%s %.15g\n" % (name, value) for name, value in zip(problem.labels(), problem.getp()))

    with open(path / filename, "wt") as pars_file:
        pars_file.write(pardata)
    return {"filename": filename, "check_overwrite": False}


@register
async def apply_parameters(pathlist: List[str], filename: str):
    path = Path(*pathlist)
    fullpath = path / filename
    try:
        bumps.cli.load_best(state.problem.fitProblem, fullpath)
        state.shared.updated_parameters = now_string()
        await log(f"Applied parameters from {fullpath}")
        await add_notification(
            f"Applied parameters from {fullpath}",
            title="Parameters applied",
            timeout=2000,
        )
    except Exception:
        await log(f"Unable to apply parameters from {fullpath}")
        await add_notification(
            f"Unable to apply parameters from {fullpath}",
            title="Error applying parameters",
            timeout=2000,
        )


@register
async def start_fit(fitter_id: str = "", kwargs=None):
    kwargs = {} if kwargs is None else kwargs
    problem_state = state.problem
    if problem_state is None:
        await log("Error: Can't start fit if no problem loaded")
    else:
        fitProblem = problem_state.fitProblem
        mapper = MPMapper.start_mapper(fitProblem, None, cpus=state.parallel)
        monitors = []
        fitclass = FITTERS_BY_ID[fitter_id]
        driver = FitDriver(
            fitclass=fitclass,
            mapper=mapper,
            problem=fitProblem,
            monitors=monitors,
            **kwargs,
        )
        x, fx = driver.fit()
        driver.show()


@register
async def stop_fit():
    if state.fit_thread is not None:
        if state.fit_thread.is_alive():
            state.fit_abort_event.set()
            loop = getattr(state, "calling_loop", None)
            if loop is not None:
                await loop.run_in_executor(None, state.fit_complete_event.wait)
    else:
        state.shared.active_fit = {}


@register
async def get_chisq(problem: Optional[bumps.fitproblem.FitProblem] = None, nllf=None):
    problem = state.problem.fitProblem if problem is None else problem
    if problem is None:
        return ""
    nllf = problem.nllf() if nllf is None else nllf
    scale, err = nllf_scale(problem)
    chisq = format_uncertainty(scale * nllf, err)
    return chisq


def get_num_steps(fitter_id: str, num_fitparams: int, options: Optional[Dict] = None):
    options = FITTER_DEFAULTS[fitter_id] if options is None else options
    steps = options["steps"]
    if fitter_id == "dream":
        if steps == 0:
            total_pop = options["pop"] * num_fitparams
            steps = int(options["samples"] / total_pop)
        steps += options["burn"]
    return steps


def get_running_loop():
    try:
        return asyncio.get_running_loop()
    except RuntimeError:
        return None


@register
async def start_fit_thread(fitter_id: str = "", options=None, terminate_on_finish=False):
    options = {} if options is None else options  # session_id: str = app["active_session"]
    fitProblem = state.problem.fitProblem if state.problem is not None else None
    if fitProblem is None:
        await log("Error: Can't start fit if no problem loaded")
    else:
        state.calling_loop = get_running_loop()
        fit_state = state.fitting
        fitclass = FITTERS_BY_ID[fitter_id]
        if state.fit_thread is not None:
            # warn that fit is alread running...
            logger.warning("fit already running...")
            await log("Can't start fit, a fit is already running...")
            return

        # TODO: better access to model parameters
        num_params = len(fitProblem.getp())
        if num_params == 0:
            raise ValueError("Problem has no fittable parameters")

        # Start a new thread worker and give fit problem to the worker.
        # Clear abort and uncertainty state
        # state.abort = False
        # state.fitting.uncertainty_state = None
        num_steps = get_num_steps(fitter_id, num_params, options)
        state.fit_abort_event.clear()
        state.fit_complete_event.clear()
        state.fit_uncertainty_final.clear()

        fit_thread = FitThread(
            abort_event=state.fit_abort_event,
            problem=fitProblem,
            fitclass=fitclass,
            options=options,
            parallel=state.parallel,
            # session_id=session_id,
            # Number of seconds between updates to the GUI, or 0 for no updates
            convergence_update=5,
            uncertainty_update=state.shared.autosave_session_interval,
            terminate_on_finish=terminate_on_finish,
        )
        state.shared.active_fit = to_json_compatible_dict(
            dict(
                fitter_id=fitter_id,
                options=options,
                num_steps=num_steps,
                step=0,
                chisq="",
                value=0,
            )
        )
        await log(
            json.dumps(to_json_compatible_dict(options), indent=2),
            title=f"Starting fitter {fitter_id}",
        )
        state.autosave()
        fit_thread.start()
        state.fit_thread = fit_thread


async def _fit_progress_handler(event: Dict):
    # session_id = event["session_id"]
    if TRACE_MEMORY:
        import tracemalloc

        if tracemalloc.is_tracing():
            snapshot = tracemalloc.take_snapshot()
            top_stats = snapshot.statistics("lineno")
            print("memory use:")
            for stat in top_stats[:15]:
                print(stat)

    problem_state = state.problem
    fitProblem = problem_state.fitProblem if problem_state is not None else None
    if fitProblem is None:
        raise ValueError("should never happen: fit progress reported for session in which fitProblem is undefined")
    message = event.get("message", None)
    if message == "complete" or message == "improvement":
        fitProblem.setp(event["point"])
        fitProblem.model_update()
        state.shared.updated_parameters = now_string()
        if message == "complete":
            state.shared.active_fit = {}
    elif message == "convergence_update":
        state.fitting.population = event["pop"]
        state.shared.updated_convergence = now_string()
    elif message == "progress":
        active_fit = state.shared.active_fit
        active_fit.update({"step": event["step"], "chisq": event["chisq"]})
        state.shared.active_fit = active_fit
    elif message == "uncertainty_update" or message == "uncertainty_final":
        state.fitting.uncertainty_state = cast(bumps.dream.state.MCMCDraw, event["uncertainty_state"])
        state.shared.updated_uncertainty = now_string()
        state.autosave()
        if message == "uncertainty_final":
            # fit is not complete until uncertainty is saved.
            state.fit_uncertainty_final.set()


async def _fit_complete_handler(event):
    message = event.get("message", None)
    fit_thread = state.fit_thread
    terminate = False
    if fit_thread is not None:
        terminate = fit_thread.terminate_on_finish
        fit_thread.join(1)  # 1 second timeout on join
        if fit_thread.is_alive():
            await log("Fit thread failed to complete")
    state.fit_thread = None
    state.shared.active_fit = {}
    if message == "error":
        await log(
            event["traceback"],
            title=f"Fit failed with error: {event.get('error_string')}",
        )
        logger.warning(f"Fit failed with error: {event.get('error_string')}\n{event['traceback']}")
    else:
        problem: bumps.fitproblem.FitProblem = event["problem"]
        chisq = nice(2 * event["value"] / problem.dof)
        problem.setp(event["point"])
        problem.model_update()
        state.problem.fitProblem = problem
        if state.shared.autosave_history:
            await save_to_history(
                f"Fit complete: {event['fitter_id']}",
                include_population=True,
                include_uncertainty=True,
            )
        state.shared.updated_parameters = now_string()
        await log(event["info"], title=f"Done with chisq {chisq}")
        logger.info(f"Fit done with chisq {chisq}")

    state.fit_complete_event.set()

    if terminate:
        await shutdown()


def fit_progress_handler(event: Dict):
    loop = getattr(state, "calling_loop", None)
    if loop is not None:
        asyncio.run_coroutine_threadsafe(_fit_progress_handler(event), loop)


def fit_complete_handler(event: Dict):
    loop = getattr(state, "calling_loop", None)
    if event["message"] != "error":
        state.fit_uncertainty_final.wait()
    if loop is not None:
        asyncio.run_coroutine_threadsafe(_fit_complete_handler(event), loop)


EVT_FIT_PROGRESS.connect(fit_progress_handler, weak=True)
EVT_FIT_COMPLETE.connect(fit_complete_handler, weak=True)


async def log(message: str, title: Optional[str] = None):
    topic = "log"
    contents = {
        "message": {"message": message, "title": title},
        "timestamp": now_string(),
    }
    # session = get_session(session_id)
    state.topics[topic].append(contents)
    # if session_id == app["active_session"]:
    await emit(topic, contents)


@register
async def get_data_plot(model_indices: Optional[List[int]] = None):
    if state.problem is None or state.problem.fitProblem is None:
        return None
    fitProblem = deepcopy(state.problem.fitProblem)
    import mpld3
    import matplotlib

    matplotlib.use("agg")
    import matplotlib.pyplot as plt
    import time

    start_time = time.time()
    logger.info(f"queueing new data plot... {start_time}")
    fig = plt.figure()
    for i, model in enumerate(fitProblem.models):
        if model_indices is not None and i not in model_indices:
            continue
        model.plot()
    plt.text(0.01, 0.01, "chisq=%s" % fitProblem.chisq_str(), transform=plt.gca().transAxes)
    dfig = mpld3.fig_to_dict(fig)
    plt.close(fig)
    end_time = time.time()
    logger.info(f"time to draw data plot: {end_time - start_time}")
    return {"fig_type": "mpld3", "plotdata": to_json_compatible_dict(dfig)}


@register
async def get_model_names():
    problem = state.problem.fitProblem
    if problem is None:
        return None
    return [p.name if p.name is not None else f"model_{i}" for (i, p) in enumerate(problem.models)]


@register
async def get_model():
    if state.problem is None or state.problem.fitProblem is None:
        return None
    fitProblem = state.problem.fitProblem
    serialized = serialize_problem(fitProblem, "dataclass") if state.problem.serializer == "dataclass" else "null"
    return serialized


class WebviewPlotFunction(Protocol):
    def __call__(
        self,
        model: bumps.fitproblem.Fitness,
        problem: bumps.fitproblem.FitProblem,
        state: Optional[bumps.dream.state.MCMCDraw] = None,
        n_samples: Optional[int] = None,
    ) -> dict: ...


# custom plots are an opt-in feature for models
# they are defined in the model file as a dictionary of functions
# with a "change_with" key that specifies whether the plot should
# change with the uncertainty state or with the parameters
@register
async def get_custom_plot_info():
    if state.problem is None or state.problem.fitProblem is None:
        return None
    fitProblem = state.problem.fitProblem
    output: List[dict] = []
    for model_index, model in enumerate(fitProblem.models):
        model_webview_plots = getattr(model, "webview_plots", {})
        for title, plotinfo in model_webview_plots.items():
            output.append(
                {
                    "model_index": model_index,
                    "change_with": plotinfo["change_with"],
                    "title": title,
                }
            )

    return output


async def create_custom_plot(model_index: int, plot_title: str, n_samples: int = 1) -> CustomWebviewPlot:
    if state.problem is None or state.problem.fitProblem is None:
        return None
    fitProblem = deepcopy(state.problem.fitProblem)
    uncertainty_state = state.fitting.uncertainty_state

    # update model
    model = list(fitProblem.models)[model_index]
    webview_plots = getattr(model, "webview_plots", {})
    plot_info = webview_plots.get(plot_title, {})
    plot_function: WebviewPlotFunction = webview_plots.get(plot_title, {}).get("func", None)
    if plot_function is not None:
        try:
            model.update()
            model.nllf()
            if plot_info.get("change_with", None) == "uncertainty":
                plot_item: CustomWebviewPlot = await asyncio.to_thread(
                    plot_function, model, fitProblem, uncertainty_state, n_samples
                )
            else:
                plot_item: CustomWebviewPlot = await asyncio.to_thread(plot_function, model, fitProblem)
        except:
            plot_item = CustomWebviewPlot(fig_type="error", plotdata=traceback.format_exc())

        return process_custom_plot(plot_item)

    return {}


@register
async def get_custom_plot(model_index: int, plot_title: str, n_samples: int = 1):
    output = CustomWebviewPlot(figtype="error", plotdata="no plot")
    if model_index is not None:
        figdict = await create_custom_plot(model_index=model_index, plot_title=plot_title, n_samples=n_samples)

    output = to_json_compatible_dict(figdict)
    return output


@register
async def get_convergence_plot():
    if state.problem is None or state.problem.fitProblem is None:
        return None
    fitProblem = state.problem.fitProblem
    population = state.fitting.population
    if population is not None:
        normalized_pop = 2 * population / fitProblem.dof
        best, pop = normalized_pop[:, 0], normalized_pop[:, 1:]

        ni, npop = pop.shape
        iternum = np.arange(1, ni + 1)
        tail = int(0.25 * ni)
        fig = {"data": [], "layout": {}}  # go.Figure()
        hovertemplate = "(%{{x}}, %{{y}})<br>{label}<extra></extra>"
        x = iternum[tail:].tolist()
        if npop == 5:
            # fig['data'].append(dict(type="scattergl", x=x, y=pop[tail:,4].tolist(), name="95%", mode="lines", line=dict(color="lightblue", width=1), showlegend=True, hovertemplate=hovertemplate.format(label="95%")))
            fig["data"].append(
                dict(
                    type="scattergl",
                    x=x,
                    y=pop[tail:, 3].tolist(),
                    mode="lines",
                    line=dict(color="lightgreen", width=0),
                    showlegend=False,
                    hovertemplate=hovertemplate.format(label="80%"),
                )
            )
            fig["data"].append(
                dict(
                    type="scattergl",
                    x=x,
                    y=pop[tail:, 1].tolist(),
                    name="20% to 80% range",
                    fill="tonexty",
                    mode="lines",
                    line=dict(color="lightgreen", width=0),
                    hovertemplate=hovertemplate.format(label="20%"),
                )
            )
            fig["data"].append(
                dict(
                    type="scattergl",
                    x=x,
                    y=pop[tail:, 2].tolist(),
                    name="population median",
                    mode="lines",
                    line=dict(color="green"),
                    opacity=0.5,
                )
            )
            fig["data"].append(
                dict(
                    type="scattergl",
                    x=x,
                    y=pop[tail:, 0].tolist(),
                    name="population best",
                    mode="lines",
                    line=dict(color="green", dash="dot"),
                )
            )

        fig["data"].append(
            dict(
                type="scattergl",
                x=x,
                y=best[tail:].tolist(),
                name="best",
                line=dict(color="red", width=1),
                mode="lines",
            )
        )
        fig["layout"].update(
            dict(
                template="simple_white",
                legend=dict(x=1, y=1, xanchor="right", yanchor="top"),
            )
        )
        fig["layout"].update(dict(title=dict(text="Convergence", xanchor="center", x=0.5)))
        fig["layout"].update(dict(uirevision=1))
        fig["layout"].update(
            dict(
                xaxis=dict(
                    title="iteration number",
                    showline=True,
                    showgrid=False,
                    zeroline=False,
                )
            )
        )
        fig["layout"].update(dict(yaxis=dict(title="chisq", showline=True, showgrid=False, zeroline=False)))
        return to_json_compatible_dict(fig)
    else:
        return None


@lru_cache(maxsize=30)
def _get_correlation_plot(
    sort: bool = True,
    max_rows: int = 8,
    nbins: int = 50,
    vars=None,
    timestamp: str = "",
):
    from .corrplot import Corr2d

    uncertainty_state = state.fitting.uncertainty_state

    if isinstance(uncertainty_state, bumps.dream.state.MCMCDraw):
        import time

        start_time = time.time()
        logger.info(f"queueing new correlation plot... {start_time}")
        draw = uncertainty_state.draw(vars=vars)
        c = Corr2d(draw.points.T, bins=nbins, labels=draw.labels)
        fig = c.plot(sort=sort, max_rows=max_rows)
        logger.info(f"time to render but not serialize... {time.time() - start_time}")
        serialized = to_json_compatible_dict(fig.to_dict())
        end_time = time.time()
        logger.info(f"time to draw correlation plot: {end_time - start_time}")
        return serialized
    else:
        return None


@register
async def get_correlation_plot(
    sort: bool = True,
    max_rows: int = 8,
    nbins: int = 50,
    vars=None,
    timestamp: str = "",
):
    # need vars to be immutable (hashable) for caching based on arguments:
    vars = tuple(vars) if vars is not None else None
    result = await asyncio.to_thread(
        _get_correlation_plot,
        sort=sort,
        max_rows=max_rows,
        nbins=nbins,
        vars=vars,
        timestamp=timestamp,
    )
    return result


@lru_cache(maxsize=30)
def _get_uncertainty_plot(timestamp: str = "", cbar_colors: int = 8):
    uncertainty_state = state.fitting.uncertainty_state
    if uncertainty_state is not None:
        import time

        start_time = time.time()
        logger.info(f"queueing new uncertainty plot... {start_time}")
        draw = uncertainty_state.draw()
        nbins = max(min(draw.points.shape[0] // 20000, 100), 30)
        stats = bumps.dream.stats.var_stats(draw)
        fig = plot_vars(draw, stats, nbins=nbins, cbar_colors=cbar_colors)
        logger.info(f"time to draw uncertainty plot: {time.time() - start_time}")
        return to_json_compatible_dict(fig)
    else:
        return None


@register
async def get_uncertainty_plot(timestamp: str = ""):
    result = await asyncio.to_thread(_get_uncertainty_plot, timestamp=timestamp, cbar_colors=8)
    return result


@register
async def get_model_uncertainty_plot():
    if state.problem is None or state.problem.fitProblem is None:
        return None
    fitProblem = state.problem.fitProblem
    uncertainty_state = state.fitting.uncertainty_state
    if uncertainty_state is not None:
        import mpld3
        import matplotlib

        matplotlib.use("agg")
        import matplotlib.pyplot as plt
        import time

        start_time = time.time()
        logger.info(f"queueing new model uncertainty plot... {start_time}")

        fig = plt.figure()
        errs = bumps.errplot.calc_errors_from_state(fitProblem, uncertainty_state)
        logger.info(f"errors calculated: {time.time() - start_time}")
        bumps.errplot.show_errors(errs, fig=fig)
        logger.info(f"time to render but not serialize... {time.time() - start_time}")
        fig.canvas.draw()
        dfig = mpld3.fig_to_dict(fig)
        plt.close(fig)
        end_time = time.time()
        logger.info(f"time to draw model uncertainty plot: {end_time - start_time}")
        return dfig
    else:
        return None


@register
async def get_parameter_labels():
    # Required to support get_parameter_trace_plot because ordering must be preserved.
    # There is no way to know whether a disambiguated name occurred first or second from
    # get_parameters. Uses fitProblem because uncertainty state might not exist and
    # list of parameters should be updated on model_loaded.
    # Should probably be able to call these parameters by ID.

    if state.problem is None or state.problem.fitProblem is None:
        return None
    return to_json_compatible_dict(state.problem.fitProblem.labels())


@register
async def get_parameter_trace_plot(var: int):
    uncertainty_state = state.fitting.uncertainty_state
    if uncertainty_state is not None:
        import time

        start_time = time.time()
        logger.info(f"queueing new parameter_trace plot... {start_time}")

        # begin plotting:
        portion = None
        draw, points, _ = uncertainty_state.chains()
        label = uncertainty_state.labels[var]
        start = int((1 - portion) * len(draw)) if portion else 0
        genid = (
            np.arange(
                uncertainty_state.generation - len(draw) + start,
                uncertainty_state.generation,
            )
            + 1
        )
        fig = plot_trace(
            genid * uncertainty_state.thinning,
            np.squeeze(points[start:, uncertainty_state._good_chains, var]).T,
            label=label,
            alpha=0.4,
        )

        logger.info(f"time to render but not serialize... {time.time() - start_time}")
        dfig = fig.to_dict()
        end_time = time.time()
        logger.info(f"time to draw parameter_trace plot: {end_time - start_time}")
        return to_json_compatible_dict(dfig)
    else:
        return None


@register
async def get_parameters(only_fittable: bool = False):
    if state.problem is None or state.problem.fitProblem is None:
        return []
    fitProblem = state.problem.fitProblem

    all_parameters = fitProblem.model_parameters()
    if only_fittable:
        parameter_infos = params_to_list(unique(all_parameters))
        # only include params with priors:
        parameter_infos = [pi for pi in parameter_infos if pi["fittable"] and not pi["fixed"]]
    else:
        parameter_infos = params_to_list(all_parameters)

    return to_json_compatible_dict(parameter_infos)


@register
async def set_parameter(
    parameter_id: str,
    property: Literal["value01", "value", "min", "max"],
    value: Union[float, str, bool],
):
    if state.problem is None or state.problem.fitProblem is None:
        return None
    fitProblem = state.problem.fitProblem

    parameter = fitProblem._parameters_by_id.get(parameter_id, None)
    if parameter is None:
        warnings.warn(f"Attempting to update parameter that doesn't exist: {parameter_id}")
        return

    if parameter.prior is None:
        warnings.warn(f"Attempting to set prior properties on parameter without priors: {parameter}")
        return

    if property == "value01":
        new_value = parameter.prior.put01(value)
        nice_new_value = nice(new_value, digits=VALUE_PRECISION)
        parameter.clip_set(nice_new_value)
    elif property == "value":
        new_value = float(value)
        nice_new_value = nice(new_value, digits=VALUE_PRECISION)
        parameter.clip_set(nice_new_value)
    elif property == "min":
        lo = float(value)
        hi = parameter.prior.limits[1]
        parameter.range(lo, hi)
        parameter.add_prior()
    elif property == "max":
        lo = parameter.prior.limits[0]
        hi = float(value)
        parameter.range(lo, hi)
        parameter.add_prior()
    elif property == "fixed":
        if parameter.fittable:
            parameter.fixed = bool(value)
            fitProblem.model_reset()
            # logger.info(f"setting parameter: {parameter}.fixed to {value}")
            # model has been changed: setp and getp will return different values!
            state.shared.updated_model = now_string()
    fitProblem.model_update()
    state.shared.updated_parameters = now_string()
    return


def now_string():
    return f"{datetime.now().timestamp():.6f}"


@register
async def publish(topic: str, message: Any = None):
    timestamp_str = f"{datetime.now().timestamp():.6f}"
    contents = {"message": message, "timestamp": timestamp_str}
    # session = get_session(session_id)
    state.topics[topic].append(contents)
    # if session_id == app["active_session"]:
    await emit(topic, contents)
    # logger.info(f"emitted: {topic} :: {contents}")


@register
async def get_shared_setting(setting: str):
    value = await state.shared.get(setting)
    return to_json_compatible_dict(value)


@register
async def set_shared_setting(setting: str, value: Any):
    await state.shared.set(setting, value)


async def notify_shared_setting(setting: str, value: Any):
    await emit(setting, to_json_compatible_dict(value))


state.shared._notification_callbacks["emit"] = notify_shared_setting


@register
async def get_topic_messages(topic: Optional[TopicNameType] = None, max_num=None) -> List[Dict]:
    # this is a GET request in disguise -
    # emitter must handle the response in a callback,
    # as no separate response event is emitted.
    if topic is None:
        return []
    topics = state.topics
    q = topics.get(topic, None)
    if q is None:
        raise ValueError(f"Topic: {topic} not defined")
    elif max_num is None:
        return list(q)
    else:
        q_length = len(q)
        start = max(q_length - max_num, 0)
        return list(itertools.islice(q, start, q_length))


@register
async def get_dirlisting(pathlist: Optional[List[str]] = None):
    # GET request
    # TODO: use psutil to get disk listing as well?
    subfolders = []
    files = []
    path = Path(state.base_path) if (pathlist is None or len(pathlist) == 0) else Path(*pathlist)
    if not path.exists():
        await add_notification(
            f"Path does not exist: {path}, falling back to current working directory",
            title="Error",
            timeout=2000,
        )
        path = Path.cwd()

    abs_path = path.absolute()
    for p in abs_path.iterdir():
        if not p.exists():
            continue
        stat = p.stat()
        mtime = stat.st_mtime
        fileinfo = {"name": p.name, "modified": mtime}
        if p.is_dir():
            fileinfo["size"] = len(list(p.glob("*")))
            subfolders.append(fileinfo)
        else:
            # files.append(p.resolve().name)
            fileinfo["size"] = stat.st_size
            files.append(fileinfo)
    # for Windows: list drives as well
    drives = os.listdrives() if hasattr(os, "listdrives") else []
    return dict(drives=drives, pathlist=abs_path.parts, subfolders=subfolders, files=files)


@register
async def get_fitter_defaults(*args):
    return FITTER_DEFAULTS


@register
async def shutdown():
    logger.info("killing...")
    await stop_fit()
    state.autosave()
    await emit("server_shutting_down")
    shutdown_result = asyncio.gather(_shutdown(), return_exceptions=True)


async def add_notification(content: str, title: str = "Notification", timeout: Optional[int] = None):
    id = None
    if timeout is None:
        id = str(uuid.uuid4())
        await emit("add_notification", {"title": title, "content": content, "id": id})
    else:
        await emit("add_notification", {"title": title, "content": content, "timeout": timeout})
    return id


async def _shutdown():
    raise SystemExit(0)


VALUE_PRECISION = 6
VALUE_FORMAT = "{{:.{:d}g}}".format(VALUE_PRECISION)


def nice(v, digits=4):
    """Fix v to a value with a given number of digits of precision"""
    from math import log10, floor

    v = float(v)
    if v == 0.0 or not np.isfinite(v):
        return v
    else:
        sign = v / abs(v)
        place = floor(log10(abs(v)))
        scale = 10 ** (place - (digits - 1))
        return sign * floor(abs(v) / scale + 0.5) * scale


JSON_TYPE = Union[str, float, bool, None, Sequence["JSON_TYPE"], Mapping[str, "JSON_TYPE"]]


def to_json_compatible_dict(obj) -> JSON_TYPE:
    if isinstance(obj, (list, tuple)):
        return type(obj)(to_json_compatible_dict(v) for v in obj)
    elif isinstance(obj, GeneratorType):
        return list(to_json_compatible_dict(v) for v in obj)
    elif isinstance(obj, dict):
        return type(obj)((to_json_compatible_dict(k), to_json_compatible_dict(v)) for k, v in obj.items())
    elif isinstance(obj, np.ndarray) and obj.dtype.kind in ["f", "i"]:
        return obj.tolist()
    elif isinstance(obj, np.ndarray) and obj.dtype.kind == "O":
        return to_json_compatible_dict(obj.tolist())
    elif isinstance(obj, bool) or isinstance(obj, str) or obj is None:
        return obj
    elif isinstance(obj, numbers.Number):
        return str(obj) if not np.isfinite(obj) else float(obj)
    elif isinstance(obj, UNDEFINED_TYPE):
        return None
    else:
        raise ValueError("obj %s is not serializable" % str(obj))


class ParamInfo(TypedDict, total=False):
    id: str
    name: str
    paths: List[str]
    value_str: str
    fittable: bool
    fixed: bool
    writable: bool
    value01: float
    min_str: str
    max_str: str


def params_to_list(params, lookup=None, pathlist=None, links=None) -> List[ParamInfo]:
    lookup: Dict[str, ParamInfo] = {} if lookup is None else lookup
    pathlist = [] if pathlist is None else pathlist
    if isinstance(params, dict):
        for k in sorted(params.keys()):
            params_to_list(params[k], lookup=lookup, pathlist=pathlist + [k])
    elif isinstance(params, tuple) or isinstance(params, list):
        for i, v in enumerate(params):
            # add index to last item in pathlist (in-place):
            new_pathlist = pathlist.copy()
            if len(pathlist) < 1:
                new_pathlist.append("")
            new_pathlist[-1] = f"{new_pathlist[-1]}[{i:d}]"
            params_to_list(v, lookup=lookup, pathlist=new_pathlist)
    elif isinstance(params, Parameter) or isinstance(params, Constant):
        path = ".".join(pathlist)
        existing = lookup.get(params.id, None)
        if existing is not None:
            existing["paths"].append(".".join(pathlist))
        else:
            value_str = VALUE_FORMAT.format(nice(params.value))
            if hasattr(params, "has_prior"):
                has_prior = params.has_prior()
            else:
                has_prior = False

            if hasattr(params, "slot"):
                writable = type(params.slot) in [Variable, Parameter]
            else:
                writable = False
            new_item: ParamInfo = {
                "id": params.id,
                "name": str(params.name),
                "paths": [path],
                "tags": getattr(params, "tags", []),
                "writable": writable,
                "value_str": value_str,
                "fittable": params.fittable,
                "fixed": params.fixed,
            }
            if has_prior:
                assert params.prior is not None
                lo, hi = params.prior.limits
                new_item["value01"] = params.prior.get01(float(params.value))
                new_item["min_str"] = VALUE_FORMAT.format(nice(lo))
                new_item["max_str"] = VALUE_FORMAT.format(nice(hi))
            lookup[params.id] = new_item
    elif callable(getattr(params, "parameters", None)):
        # handle Expression, Constant, etc.
        subparams = params.parameters()
        params_to_list(subparams, lookup=lookup, pathlist=pathlist)
    return list(lookup.values())



# This program is public domain
# Author: Paul Kienzle
"""
Conditional expression manager

Conditional expressions are formed from combinations of
and (&), or (|), exclusive or (^) and not  (!), as well
as primitive tests.   The resulting expressions can be
evaluated on particular inputs, with the inputs passed
down to the primitive tests.  Primitive tests should be
subclassed from Condition, with the __call__  method
defined to return the truth value of the condition
given the inputs.

For example, the following creates a test whether a
number is in the open interval (0,1)::

    >>> class gt(Condition):
    ...    def __init__(self, base): self.base = base
    ...    def __call__(self, test): return test>self.base
    ...    def __str__(self): return "x>%g"%self.base
    >>> class lt(Condition):
    ...    def __init__(self, base): self.base = base
    ...    def __call__(self, test): return test<self.base
    ...    def __str__(self): return "x<%g"%self.base
    >>> test = lt(1) & gt(0)
    >>> print(test)
    (x<1 and x>0)
    >>> test(0.5)
    True
    >>> test(1)
    False

No assumptions are made about the structure of the arguments
to the primitives, but all primitives in an expression should
accept the same arguments.

The constants true and false are predefined as primitives
which can take any arguments::

    >>> true()
    True
    >>> false("this",string="ignored")
    False

You can find the individual terms of the expression using
the method primitives:

    >>> (true&false).primitives() == set([true,false])
    True

In many instances you not only want to know that an expression
is true or false, but why it is true or false. The status method
on expressions does this, returning not only the truth status as
a boolean, but also a list of primitives and Not(primitives)
saying which conditions are responsible.  For example::

    >>> class iters(Condition):
    ...    def __init__(self, base): self.base = base
    ...    def __call__(self, state): return state['iters']>self.base
    ...    def __str__(self): return "iters>%d"%self.base
    >>> class stepsize(Condition):
    ...    def __init__(self, base): self.base = base
    ...    def __call__(self, state): return state['stepsize']<self.base
    ...    def __str__(self): return "stepsize<%g"%self.base
    >>> converge = stepsize(0.001)
    >>> diverge = iters(100)
    >>> result,why = converge.status(dict(stepsize=21.2,iters=20))
    >>> print("%s %s"%(result, ", ".join(str(c) for c in why)))
    False not stepsize<0.001
    >>> result,why = diverge.status(dict(stepsize=21.2,iters=129))
    >>> print("%s %s"%(result, ", ".join(str(c) for c in why)))
    True iters>100

Note that status will be less efficient than direct evaluation
because it has to test all branches and construct the result list.
Normally the And and Or calculations an short circuit, and only
compute what they need to guarantee the resulting truth value.
"""


class Condition(object):
    """
    Condition abstract base class.

    Subclasses should define __call__(self, *args, **kw)
    which returns True if the condition is satisfied, and
    False otherwise.
    """

    def __and__(self, condition):
        return And(self, condition)

    def __or__(self, condition):
        return Or(self, condition)

    def __xor__(self, condition):
        return Xor(self, condition)

    def __invert__(self):
        return Not(self)

    def __call__(self, *args, **kw):
        raise NotImplementedError

    def _negate(self):
        return _Bar(self)

    def status(self, *args, **kw):
        """
        Evaluate the condition, returning both the status and a list of
        conditions.  If the status is true, then the conditions are those
        that contribute to the true status.  If the status is false, then
        the conditions are those that contribute to the false status.
        """
        stat = self.__call__(*args, **kw)
        if stat:
            return stat, [self]
        else:
            return stat, [Not(self)]

    def primitives(self):
        """
        Return a list of terms in the condition.
        """
        return set([self])


class And(Condition):
    """
    True if both conditions are satisfied.
    """

    def __init__(self, left, right):
        self.left, self.right = left, right

    def __call__(self, *args, **kw):
        return self.left(*args, **kw) and self.right(*args, **kw)

    def __str__(self):
        return "(%s and %s)" % (self.left, self.right)

    def _negate(self):
        return Or(self.left._negate(), self.right._negate())

    def status(self, *args, **kw):
        lstat, lcond = self.left.status(*args, **kw)
        rstat, rcond = self.right.status(*args, **kw)
        if lstat and rstat:
            return True, lcond + rcond
        elif lstat:
            return False, rcond
        elif rstat:
            return False, lcond
        else:
            return False, lcond + rcond

    def primitives(self):
        return self.left.primitives() | self.right.primitives()


class Or(Condition):
    """
    True if either condition is satisfied
    """

    def __init__(self, left, right):
        self.left, self.right = left, right

    def __call__(self, *args, **kw):
        return self.left(*args, **kw) or self.right(*args, **kw)

    def __str__(self):
        return "(%s or %s)" % (self.left, self.right)

    def _negate(self):
        return And(self.left._negate(), self.right._negate())

    def status(self, *args, **kw):
        lstat, lcond = self.left.status(*args, **kw)
        rstat, rcond = self.right.status(*args, **kw)
        if lstat and rstat:
            return True, lcond + rcond
        elif lstat:
            return True, lcond
        elif rstat:
            return True, rcond
        else:
            return False, lcond + rcond

    def primitives(self):
        return self.left.primitives() | self.right.primitives()


class Xor(Condition):
    """
    True if only one condition is satisfied
    """

    def __init__(self, left, right):
        self.left, self.right = left, right

    def __call__(self, *args, **kw):
        l, r = self.left(*args, **kw), self.right(*args, **kw)
        return (l or r) and not (l and r)

    def __str__(self):
        return "(%s xor %s)" % (self.left, self.right)

    def _negate(self):
        return Xor(self.left, self.right._negate())

    def status(self, *args, **kw):
        lstat, lcond = self.left.status(*args, **kw)
        rstat, rcond = self.right.status(*args, **kw)
        if lstat ^ rstat:
            return True, lcond + rcond
        else:
            return False, lcond + rcond

    def primitives(self):
        return self.left.primitives() | self.right.primitives()


class Not(Condition):
    """
    True if condition is not satisfied
    """

    def __init__(self, condition):
        self.condition = condition

    def __call__(self, *args, **kw):
        return not self.condition(*args, **kw)

    def __str__(self):
        return "not " + str(self.condition)

    def _negate(self):
        return self.condition

    def status(self, *args, **kw):
        stat, cond = self.condition.status()
        return not stat, cond

    def __eq__(self, other):
        return isinstance(other, Not) and self.condition == other.condition

    def __ne__(self, other):
        return not isinstance(other, Not) or self.condition != other.condition

    def primitives(self):
        return self.condition.primitives()


class _Bar(Condition):
    """
    This is an internal condition structure created solely to handle
    negated primitives when computing status.  It should not be used
    externally.
    """

    def __init__(self, condition):
        self.condition = condition

    def __call__(self, *args, **kw):
        return not self.condition(*args, **kw)

    def _negate(self):
        return self.condition

    def status(self, *args, **kw):
        stat, cond = self.condition.status(*args, **kw)
        return not stat, cond

    def __str__(self):
        return "not " + str(self.condition)

    def primitives(self):
        return self.condition.primitives()


class Constant(Condition):
    """
    Constants true and false.
    """

    def __init__(self, value):
        self._value = value

    def __call__(self, *args, **kw):
        return self._value

    def __str__(self):
        return str(self._value)


true = Constant(True)
false = Constant(False)



"""
Miscellaneous utility functions.
"""

__all__ = ["kbhit", "profile", "pushdir", "push_seed", "redirect_console"]

import os
import sys
import types
from dataclasses import Field, dataclass, field, fields, is_dataclass
from io import StringIO

# this can be substituted with pydantic dataclass for schema-building...
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    runtime_checkable,
)

import numpy as np
from numpy import ascontiguousarray as _dense

# **DEPRECATED** we can import erf directly from scipy.special.erf
# so there is no longer a need for bumps.util.erf.
from scipy.special import erf

USE_PYDANTIC = os.environ.get("BUMPS_USE_PYDANTIC", "False") == "True"
if TYPE_CHECKING:
    from numpy.typing import NDArray


def field_desc(description: str) -> Any:
    return field(metadata={"description": description})


T = TypeVar("T")
SCHEMA_ATTRIBUTE_NAME = "__bumps_schema__"


def schema_config(
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
) -> Callable[[Type[T]], Type[T]]:
    """
    Add an attribute for configuring serialization
    (presence of SCHEMA_ATTRIBUTE_NAME attribute is used during
    serialization of unique instances e.g. Parameter)
    """

    def add_schema_config(cls: Type[T]) -> Type[T]:
        if include is not None and exclude is not None:
            fqn = f"{cls.__module__}.{cls.__qualname__}"
            raise ValueError(f"{fqn} schema: include array and exclude array are mutually exclusive - only define one")

        setattr(cls, SCHEMA_ATTRIBUTE_NAME, dict(include=include, exclude=exclude))
        return cls

    return add_schema_config


def has_schema_config(cls):
    return is_dataclass(cls) and hasattr(cls, SCHEMA_ATTRIBUTE_NAME)


@dataclass(init=True)
class NumpyArray:
    """
    Wrapper for numpy arrays:
    on serialize,
     - array.tolist() is called and stored in 'values' attribute
     - str(array.dtype) is stored in 'dtype' attribute

     on deserialize,
      - return new np.ndarray(values, dtype=dtype)
    """

    dtype: str
    values: Sequence = field(default_factory=list)


if USE_PYDANTIC:
    from typing import TypeAlias

    NDArray: TypeAlias = NumpyArray


def parse_errfile(errfile):
    """
    Parse dream statistics from a particular fit.

    Returns overall chisq, list of chisq for individual models and
    a parameter dictionary with attributes for number, name, mean, median,
    p68 for 68% credible interval and p95 for 95% credible interval.

    The parameter dictionary is keyed by parameter name.

    Usually there is only one errfile in a directory, which can be
    retrieved using::

        import os.path
        import glob
        errfile = glob.glob(os.path.join(path, '*.err'))[0]
    """
    from .dream.stats import parse_var

    pars = []
    chisq = []
    overall = None
    with open(errfile) as fid:
        for line in fid:
            if line.startswith("[overall"):
                overall = float(line.split()[1][6:-1])
                continue

            if line.startswith("[chisq"):
                chisq.append(float(line.split()[0][7:-1]))
                continue

            p = parse_var(line)
            if p is not None:
                pars.append(p)

    if overall is None:
        overall = chisq[0]
    pardict = dict((p.name, p) for p in pars)
    return overall, chisq, pardict


def profile(fn, *args, **kw):
    """
    Profile a function called with the given arguments.
    """
    import cProfile
    import pstats

    result = [None]

    def call():
        result[0] = fn(*args, **kw)

    datafile = "profile.out"
    cProfile.runctx("call()", dict(call=call), {}, datafile)
    stats = pstats.Stats(datafile)
    # order='calls'
    order = "cumulative"
    # order='pcalls'
    # order='time'
    stats.sort_stats(order)
    stats.print_stats()
    os.unlink(datafile)
    return result[0]


def kbhit():
    """
    Check whether a key has been pressed on the console.
    """
    try:  # Windows
        import msvcrt

        return msvcrt.kbhit()
    except ImportError:  # Unix
        import select

        i, _, _ = select.select([sys.stdin], [], [], 0.0001)
        return sys.stdin in i


class DynamicModule(types.ModuleType):
    def __init__(self, path, name):
        self.__path__ = [path]
        self.__name__ = name
        # In the bowels of importlib the parent spec attribute is used to
        # avoid circular imports, but only if spec is not None. This behaviour
        # was observed on python 3.11, and perhaps earlier.
        self.__spec__ = None


def relative_import(filename, module_name="relative_import"):
    """
    Define an empty module allowing relative imports from a script.

    By setting :code:`__package__ = relative_import(__file__)` at the top of
    your script file you can even run your model as a python script.  So long
    as the script behaviour is isolated in :code:`if __name__ == "__main__":`
    code block and :code:`problem = FitProblem(...)` is defined, the same model
    can be used both within and outside of bumps.
    """
    path = os.path.dirname(os.path.abspath(filename))
    if module_name in sys.modules and not isinstance(sys.modules[module_name], DynamicModule):
        raise ImportError("relative import would override the existing module %s. Use another name" % module_name)
    sys.modules[module_name] = DynamicModule(path, module_name)
    return module_name


class redirect_console(object):
    """
    Console output redirection context

    The output can be redirected to a string, to an already opened file
    (anything with a *write* attribute), or to a filename which will be
    opened for the duration of the with context.  Unless *stderr* is
    specified, then both standard output and standard error are
    redirected to the same file.  The open file handle is returned on
    enter, and (if it was not an already opened file) it is closed on exit.

    If no file is specified, then output is redirected to a StringIO
    object, which has a getvalue() method which can retrieve the string.
    The StringIO object is deleted when the context ends, so be sure to
    retrieve its value within the redirect_console context.

    :Example:

    Show that output is captured in a file:

        >>> from bumps.util import redirect_console
        >>> print("hello")
        hello
        >>> with redirect_console("redirect_out.log"):
        ...     print("captured")
        >>> print("hello")
        hello
        >>> print(open("redirect_out.log").read()[:-1])
        captured
        >>> import os; os.unlink("redirect_out.log")

    Output can also be captured to a string:

        >>> with redirect_console() as fid:
        ...    print("captured to string")
        ...    captured_string = fid.getvalue()
        >>> print(captured_string.strip())
        captured to string

    """

    def __init__(self, stdout=None, stderr=None):
        self.open_files = []
        self.sys_stdout = []
        self.sys_stderr = []

        if stdout is None:
            self.open_files.append(StringIO())
            self.stdout = self.open_files[-1]
        elif hasattr(stdout, "write"):
            self.stdout = stdout
        else:
            self.open_files.append(open(stdout, "w"))
            self.stdout = self.open_files[-1]

        if stderr is None:
            self.stderr = self.stdout
        elif hasattr(stderr, "write"):
            self.stderr = stderr
        else:
            self.open_files.append(open(stderr, "w"))
            self.stderr = self.open_files[-1]

    def __del__(self):
        for f in self.open_files:
            f.close()

    def __enter__(self):
        self.sys_stdout.append(sys.stdout)
        self.sys_stderr.append(sys.stderr)
        sys.stdout = self.stdout
        sys.stderr = self.stderr
        return self.open_files[-1]

    def __exit__(self, *args):
        sys.stdout = self.sys_stdout[-1]
        sys.stderr = self.sys_stderr[-1]
        del self.sys_stdout[-1]
        del self.sys_stderr[-1]
        return False


class push_python_path(object):
    """
    Change sys.path for the duration of a with statement.

    :Example:

    Show that the original directory is restored::

        >>> import sys, os
        >>> original_path = list(sys.path)
        >>> with push_python_path('/tmp'):
        ...     assert sys.path[-1] == '/tmp'
        >>> restored_path = list(sys.path)
        >>> assert original_path == restored_path
    """

    def __init__(self, path):
        self.path = path

    def __enter__(self):
        sys.path.append(self.path)

    def __exit__(self, *args):
        del sys.path[-1]


class pushdir(object):
    """
    Change directories for the duration of a with statement.

    :Example:

    Show that the original directory is restored::

        >>> import sys, os
        >>> original_wd = os.getcwd()
        >>> with pushdir(sys.path[0]):
        ...     pushed_wd = os.getcwd()
        ...     first_site = os.path.abspath(sys.path[0])
        ...     assert pushed_wd == first_site
        >>> restored_wd = os.getcwd()
        >>> assert original_wd == restored_wd
    """

    def __init__(self, path):
        self.path = os.path.abspath(path)

    def __enter__(self):
        self._cwd = os.getcwd()
        os.chdir(self.path)

    def __exit__(self, *args):
        os.chdir(self._cwd)


class push_seed(object):
    """
    Set the seed value for the random number generator.

    When used in a with statement, the random number generator state is
    restored after the with statement is complete.

    :Parameters:

    *seed* : int or array_like, optional
        Seed for RandomState

    :Example:

    Seed can be used directly to set the seed::

        >>> from numpy.random import randint
        >>> push_seed(24)
        <...push_seed object at...>
        >>> print(randint(0,1000000,3))
        [242082    899 211136]

    Seed can also be used in a with statement, which sets the random
    number generator state for the enclosed computations and restores
    it to the previous state on completion::

        >>> with push_seed(24):
        ...    print(randint(0,1000000,3))
        [242082    899 211136]

    Using nested contexts, we can demonstrate that state is indeed
    restored after the block completes::

        >>> with push_seed(24):
        ...    print(randint(0,1000000))
        ...    with push_seed(24):
        ...        print(randint(0,1000000,3))
        ...    print(randint(0,1000000))
        242082
        [242082    899 211136]
        899

    The restore step is protected against exceptions in the block::

        >>> with push_seed(24):
        ...    print(randint(0,1000000))
        ...    try:
        ...        with push_seed(24):
        ...            print(randint(0,1000000,3))
        ...            raise Exception()
        ...    except Exception:
        ...        print("Exception raised")
        ...    print(randint(0,1000000))
        242082
        [242082    899 211136]
        Exception raised
        899
    """

    def __init__(self, seed=None):
        self._state = np.random.get_state()
        np.random.seed(seed)

    def __enter__(self):
        return None

    def __exit__(self, *args):
        np.random.set_state(self._state)


def get_libraries(obj, libraries=None):
    if libraries is None:
        libraries = {}

    if is_dataclass(obj):
        _add_to_libraries(obj, libraries)
        for f in fields(obj):
            subobj = getattr(obj, f.name, None)
            if subobj is not None:
                get_libraries(subobj, libraries)
    elif isinstance(obj, (list, tuple, types.GeneratorType)):
        for v in obj:
            get_libraries(v, libraries)
    elif isinstance(obj, dict):
        for v in obj.values():
            if is_dataclass(v):
                get_libraries(v, libraries)

    return libraries


def _add_to_libraries(obj, libraries: dict):
    top_level_package = obj.__module__.split(".")[0]
    if top_level_package is not None and top_level_package not in libraries:
        # module is already in sys.modules for obj
        # get __version__ attribute from module directly
        version = getattr(sys.modules[top_level_package], "__version__", None)
        schema_version = getattr(sys.modules[top_level_package], "__schema_version__", None)
        libraries[top_level_package] = dict(version=version, schema_version=schema_version)



"""
Random Lines Algorithm finds the optimal minimum of a function.

Sahin, I. (2013). Minimization over randomly selected lines.  An
International  Journal Of Optimization And Control: Theories &
Applications (IJOCTA), 3(2), 111-119.
http://dx.doi.org/10.11121/ijocta.01.2013.00167
"""

# Author : Ismet Sahin

__all__ = ["random_lines", "particle_swarm"]

from itertools import count

from numpy import zeros, ones, asarray, sqrt, arange, isfinite
from numpy.random import rand, random_integers


def print_every_five(step, x, fx, k):
    if step % 5 == 0:
        print(step, ":", fx[k], x[k])


def random_lines(cfo, NP, CR=0.9, epsilon=1e-10, abort_test=None, maxiter=1000):
    """
    Random lines is a population based optimizer which using quadratic
    fits along randomly oriented directions.

    *cfo* is the cost function object.  This is a dictionary which contains
    the following keys:

        *cost* is the function to be optimized.  If *parallel_cost* exists,
        it should accept a list of points, not just a single point on each
        evaluation.

        *n* is the problem dimension

        *x0* is the initial point

        *x1* and *x2* are lower and upper bounds for each parameter

        *monitor* is a callable which is called each iteration using
        *callback(step, x, fx, k)*, where *step* is the iteration number,
        *x* is the population, *fx* is value of the cost function for each
        member of the population and *k* is the index of the best point in
        the population.

        *f_opt* is the target value of the optimization

    *NP* is the number of fit parameters

    *CR* is the cross-over ratio, which is the proportion of dimensions
    that participate in any random orientation vector.

    *epsilon* is the convergence criterion.

    *abort_test* is a callable which indicates whether an external processes
    requests the fit to stop.

    *maxiter* is the maximum number of generations

    Returns success, num_evals, f(x_best), x_best.
    """
    if "parallel_cost" in cfo:
        mapper = lambda v: asarray(cfo["parallel_cost"](v.T), "d")
    else:
        mapper = lambda v: asarray(list(map(cfo["cost"], v.T)), "d")
    monitor = cfo.get("monitor", print_every_five)

    n = cfo["n"]

    X = rand(n, NP)  # will hold original vectors

    # CREATE FIRST GENERATION WITH LEGAL PARAMETER VALUES AND EVALUATE COSTS
    # m th member of the population
    for m in range(0, NP):
        X[:, m] = cfo["x1"] + (cfo["x2"] - cfo["x1"]) * X[:, m]
    if "x0" in cfo:
        X[:, 0] = cfo["x0"]
    f = mapper(X)

    n_feval = NP
    f_best, i_best = min(zip(f, count()))

    # CHECK INITIAL STOPPING CRITERIA
    if abs(cfo["f_opt"] - f_best) < epsilon:
        satisfied_sc = 1
        x_best = X[:, i_best]
        return satisfied_sc, n_feval, f_best, x_best

    for L in range(1, maxiter + 1):
        # finding destination vector
        i_Xj = random_integers(0, NP - 2, NP)
        i_ge = i_Xj >= arange(0, NP)
        i_Xj[i_ge] += 1

        # choosing muk
        muk = 0.01 + 0.49 * rand(NP)
        inx = rand(NP) < 0.5
        muk[inx] = -muk[inx]

        # find xk and fk s
        Xi = X
        Xj = X[:, i_Xj]
        P = Xj - Xi
        Xk = Xi + (ones((n, 1)) * muk) * P
        fk = mapper(Xk)
        n_feval = n_feval + NP

        # find quadratic models
        if any(muk == 0) or any(muk == 1):
            satisfied_sc = 0
            x_best = X[:, i_best]
            print("muk cannot be zero or one !!!")
            return satisfied_sc, n_feval, f_best, x_best

        fi = f
        fj = f[i_Xj]
        b = (muk / (muk - 1)) * fj - ((muk + 1) / muk) * fi - (1 / (muk * (muk - 1))) * fk
        a = fj - fi - b

        crossovers = []
        for k in range(0, NP):
            if abs(a[k]) < 1e-30 or (a[k] < 0 and fk[k] > fi[k] and fk[k] > fj[k]) or not isfinite(a[k]):
                # xi survives
                continue
            else:
                # xi may not survive
                mustar = -b[k] / (2 * a[k])
                xstar = Xi[:, k] + mustar * P[:, k]

                # choosing random numbers for crossover
                rn = rand(n)
                indi = rn < 0.5 * (1 - CR)
                indj = rn > 0.5 * (1 + CR)
                xstar[indi] = Xi[indi, k]
                xstar[indj] = Xj[indj, k]

                # map into feasible set
                inx = xstar < cfo["x1"]
                xstar[inx] = cfo["x1"][inx]
                inx = xstar > cfo["x2"]
                xstar[inx] = cfo["x2"][inx]

                crossovers.append((k, xstar))

        if len(crossovers) > 0:
            idx, xstar = [asarray(v) for v in zip(*crossovers)]
            fstar = mapper(xstar.T)
            n_feval += len(crossovers)

            # xi does not survive, xstar replaces it
            update = fstar < fi[idx]
            f[idx[update]] = fstar[update]
            X[:, idx[update]] = xstar[update, :].T

        # CHECKING STOPPING CRITERIA
        f_best, i_best = min(zip(f, count()))
        if abs(cfo["f_opt"] - f_best) < epsilon:
            satisfied_sc = 1
            x_best = X[:, i_best]
            return satisfied_sc, n_feval, f_best, x_best
        if abort_test():
            break

        monitor(L, X, f, i_best)

    return 1, n_feval, f_best, X[:, i_best]


def particle_swarm(cfo, NP, epsilon=1e-10, maxiter=1000):
    """
    Particle swarm is a population based optimizer which uses force and
    momentum to select candidate points.

    *cfo* is the cost function object.  This is a dictionary which contains
    the following keys:

        *cost* is the function to be optimized.  If *parallel_cost* exists,
        it should accept a list of points, not just a single point on each
        evaluation.

        *n* is the problem dimension

        *x0* is the initial point

        *x1* and *x2* are lower and upper bounds for each parameter

        *monitor* is a callable which is called each iteration using
        *callback(step, x, fx, k)*, where *step* is the iteration number,
        *x* is the population, *fx* is value of the cost function for each
        member of the population and *k* is the index of the best point in
        the population.

        *f_opt* is the target value of the optimization

    *NP* is the number of fit parameters

    *epsilon* is the convergence criterion.

    *abort_test* is a callable which indicates whether an external processes
    requests the fit to stop.

    *maxiter* is the maximum number of generations

    Returns success, num_evals, f(x_best), x_best.
    """

    if "parallel_cost" in cfo:
        mapper = lambda v: asarray(cfo["parallel_cost"](v.T), "d")
    else:
        mapper = lambda v: asarray(list(map(cfo["cost"], v.T)), "d")
    monitor = cfo.get("monitor", print_every_five)

    n = cfo["n"]
    c1 = 2.8
    c2 = 1.3
    phi = c1 + c2
    K = 2 / abs(2 - phi - sqrt(phi * phi - 4 * phi))

    X = rand(n, NP)  # will hold original vectors
    V = zeros((n, NP))

    # CREATE FIRST GENERATION WITH LEGAL PARAMETER VALUES AND EVALUATE COSTS
    rn1 = rand(n, NP)
    # m th member of the population
    for m in range(0, NP):
        extend = cfo["x2"] - cfo["x1"]
        X[:, m] = cfo["x1"] + extend * X[:, m]
        V[:, m] = 2 * rn1[:, m] * extend - extend

    if "x0" in cfo:
        X[:, 0] = cfo["x0"]
    f = mapper(X)

    n_feval = NP
    P = X[:]

    f_best, i_best = min(zip(f, count()))
    for L in range(2, maxiter + 1):
        rn2 = rand(n, NP)
        for i in range(0, NP):
            # r = rand(2)
            r = rn2[:, i]
            V[:, i] = V[:, i] + r[0] * c1 * (P[:, i] - X[:, i]) + r[1] * c2 * (P[:, i_best] - X[:, i])
            V[:, i] = K * V[:, i]

            X[:, i] = X[:, i] + V[:, i]

        f_temp = mapper(X)
        idx = f_temp < f
        f[idx] = f_temp[idx]
        P[:, idx] = X[:, idx]

        n_feval = n_feval + NP

        # CHECKING STOPPING CRITERIA
        f_best, i_best = min(zip(f, count()))
        if abs(cfo["f_opt"] - f_best) < epsilon:
            satisfied_sc = 1
            return satisfied_sc, n_feval, f_best, X[:, i_best]

        monitor(L, X, f, i_best)

    satisfied_sc = 0
    return satisfied_sc, n_feval, f_best, X[:, i_best]


def example_call(optimizer=random_lines):
    from numpy.random import seed

    seed(1)
    cost = lambda x: x[0] ** 2 + x[1] ** 2
    n = 2
    x1 = -5 * ones(n)
    x2 = 5 * ones(n)
    f_opt = 0
    cfo = {"cost": cost, "n": n, "x1": x1, "x2": x2, "f_opt": f_opt}

    NP = 10 * n
    satisfied_sc, n_feval, f_best, x_best = optimizer(cfo, NP)
    print(satisfied_sc, "n:%d" % n_feval, f_best, x_best)


def main():
    print("=== Random Lines")
    example_call(random_lines)
    print("=== Particle Swarm")
    example_call(particle_swarm)


if __name__ == "__main__":
    main()



from os.path import join as joinpath, realpath, dirname, exists
from ctypes import CDLL

_dll_path = joinpath(realpath(dirname(__file__)), "_compiled.so")
dll = CDLL(_dll_path) if exists(_dll_path) else None



# This program is in the public domain
# Author: Paul Kienzle
"""
Bumps: curve fitter with uncertainty estimation

This package provides tools for modeling parametric systems in a Bayesian
context, with routines for finding the maximum likelihood and the
posterior probability density function.

A graphical interface allows direct manipulation of the model parameters.

See https://bumps.readthedocs.io for online manuals.
"""

try:
    from ._version import __version__  # noqa: F401
except ImportError:
    __version__ = "unknown"

__schema_version__ = "1"


def data_files():
    """
    Return the data files associated with the package for setup_py2exe.py.

    The format is a list of (directory, [files...]) pairs which can be
    used directly in the py2exe setup script as::

        setup(...,
              data_files=data_files(),
              ...)
    """
    from .gui.utilities import data_files

    return data_files()


def package_data():
    """
    Return the data files associated with the package for setup.py.

    The format is a dictionary of {'fully.qualified.module', [files...]}
    used directly in the setup script as::

        setup(...,
              package_data=package_data(),
              ...)
    """
    from .gui.utilities import package_data

    return package_data()



# This code is in the public domain.
# choose_without_replacement is copyright (2007) Anne Archibald
"""
Utility functions

choose_without_replacement(m,n,repeats=None)

    sample from a set without replacement

runlength(v)

    return the values and lengths of runs in a vector

countunique(A)

    return the element and frequency of each unique value in an array
"""

import numpy as np


# Author: Anne Archibald
# Re: [Numpy-discussion] Generating random samples without repeats
# Fri, 19 Sep 2008 12:19:22 -0700
def choose_without_replacement(m, n, repeats=None):
    """
    Choose n nonnegative integers less than m without replacement

    Returns an array of shape n, or (n,repeats).
    """
    if repeats is None:
        r = 1
    else:
        r = repeats
    if n > m:
        raise ValueError("Cannot find %d nonnegative integers less than %d" % (n, m))
    elif n > m / 2:
        res = np.sort(np.random.rand(m, r).argsort(axis=0)[:n, :], axis=0)
    else:
        res = np.random.random_integers(0, m - 1, size=(n, r))
        while True:
            res = np.sort(res, axis=0)
            w = np.nonzero(np.diff(res, axis=0) == 0)
            nr = len(w[0])
            if nr == 0:
                break
            res[w] = np.random.random_integers(0, m - 1, size=nr)

    if repeats is None:
        return res[:, 0]
    else:
        return res


def runlength(v):
    """
    Return the run lengths for repeated values in a vector v.

    See also countunique.
    """
    if len(v) == 0:
        return [], []
    diffs = np.diff(v)
    steps = np.where(diffs != 0)[0] + 1
    ends = np.hstack([[0], steps, [len(v)]])
    vals = v[ends[:-1]]
    lens = np.diff(ends)
    return vals, lens


def countunique(A):
    """
    Returns the unique elements in an array and their frequency.
    """
    return runlength(np.sort(A.flatten()))


def zscore(A, axis=None):
    """
    Convert an array of data to zscores.

    Use *axis* to limit the calculation of mean and standard deviation to
    a particular axis.
    """
    return (A - np.mean(A, axis=axis)) / np.std(A, axis=axis, ddof=1)



# This program is in the public domain
"""
Progress monitors.

Process monitors accept a :class:`bumps.history.History` object each cycle
and perform some sort of work.

Monitors have a :meth:`Monitor.config_history` method which calls
*history.requires()* to set the amount of history it needs and a
*Monitor.__call__* method which takes the updated history and
generates the monitor output.

Most monitors are subclassed from :class:`TimedUpdate` to set a minimum
time between updates and to only show updates when there is an improvement.
The *TimedUpdate* subclasses must override :meth:`TimedUpdate.show_progress`
and :meth:`TimedUpdate.show_improvement` to control the output form. History
must be updated with time, value, point and step. The
bumps.fitters.MonitorRunner class manages history and updates.
"""

__all__ = ["Monitor", "Logger", "TimedUpdate"]

from numpy import inf


class Monitor(object):
    """
    Base class for monitors.
    """

    def config_history(self, history):
        """
        Indicate which fields are needed by the monitor and for what duration.
        """
        pass

    def __call__(self, history):
        """
        Give the monitor a new piece of history to work with.
        """
        pass


def _getfield(history, field):
    """
    Return the last value in the trace, or None if there is no
    last value or no trace.
    """
    trace = getattr(history, field, [])
    try:
        return trace[0]
    except IndexError:
        return None


class Logger(Monitor):
    """
    Keeps a record of all values for the desired fields.

    *fields* is a list of history fields to store.

    *table* is an object with a *store(field=value,...)* method, which gets
    the current value of each field every time the history is updated.

    Call :meth:`config_history` with the :class:`bumps.history.History`
    object before starting so that the correct fields are stored.
    """

    def __init__(self, fields=(), table=None):
        self.fields = fields
        self.table = table

    def config_history(self, history):
        """
        Make sure history records each logged field.
        """
        kwargs = dict((key, 1) for key in self.fields)
        history.requires(**kwargs)

    def __call__(self, history):
        """
        Record the next piece of history.
        """
        record = dict((f, _getfield(history, f)) for f in self.fields)
        self.table.store(**record)


class TimedUpdate(Monitor):
    """
    Indicate progress every n seconds.

    The process should provide time, value, point, and step to the
    history update. Call :meth:`config_history` with the
    :class:`bumps.history.History` object before starting so that
    these fields are stored.

    *progress* is the number of seconds to go before showing progress, such
    as time or step number.

    *improvement* is the number of seconds to go before showing
    improvements to value.

    By default, the update only prints step number and improved value.
    Subclass TimedUpdate with replaced :meth:`show_progress` and
    :meth:`show_improvement` to trigger GUI updates or show parameter
    values.
    """

    def __init__(self, progress=60, improvement=5):
        self.progress_delta = progress
        self.improvement_delta = improvement
        self.progress_time = -inf
        self.improvement_time = -inf
        self.value = inf
        self.improved = False

    def config_history(self, history):
        history.requires(time=1, value=1, point=1, step=1)

    def show_improvement(self, history):
        print("step", history.step, "value", history.value)

    def show_progress(self, history):
        pass

    def __call__(self, history):
        t = history.time[0]
        v = history.value[0]
        if v < self.value:
            self.improved = True
        if t > self.progress_time + self.progress_delta:
            self.progress_time = t
            self.show_progress(history)
        if self.improved and t > self.improvement_time + self.improvement_delta:
            self.improved = False
            self.improvement_time = t
            self.show_improvement(history)



"""
Population initialization strategies.

To start the analysis an initial population is required.  This will be
an array of size M x N, where M is the number of dimensions in the fitting
problem and N is the number of individuals in the population.

Normally the initialization will use a call to :func:`generate` with
key-value pairs from the command line options.  This will include the
'init' option, with the name of the strategy used to initialize the
population.

Additional strategies like uniform box in [0,1] or standard norm
(rand(m,n) and randn(m,n) respectively), may also be useful.
"""

# Note: borrowed from DREAM and extended.

__all__ = ["generate", "cov_init", "eps_init", "lhs_init", "random_init"]

import math
import numpy as np
from numpy import diag, empty, isinf, isfinite, clip, inf

try:
    from typing import Optional
except ImportError:
    pass


def generate(problem, init="eps", pop=10, use_point=True, **options):
    # type: (Any, str, int, bool, ...) -> np.ndarray
    """
    Population initializer.

    *problem* is a fit problem with *getp* and *bounds* methods.

    *init* is 'eps', 'cov', 'lhs' or 'random', indicating which
    initializer should be used.

    *pop* is the population scale factor, generating *pop* individuals
    for each parameter in the fit. If *pop < 0*, generate a total of
    *-pop* individuals regardless of the number of parameters.

    *use_point* is True if the initial value should be a member of the
    population.

    Additional options are ignored so that generate can be called using
    all command line options.
    """
    initial = problem.getp()
    initial[~isfinite(initial)] = 1.0
    pop_size = int(math.ceil(pop * len(initial))) if pop > 0 else int(-pop)
    bounds = problem.bounds()
    if init == "random":
        population = random_init(pop_size, initial, bounds, use_point=use_point, problem=problem)
    elif init == "cov":
        cov = problem.cov()
        population = cov_init(pop_size, initial, bounds, use_point=use_point, cov=cov)
    elif init == "lhs":
        population = lhs_init(pop_size, initial, bounds, use_point=use_point)
    elif init == "eps":
        population = eps_init(pop_size, initial, bounds, use_point=use_point, eps=1e-6)
    else:
        raise ValueError("Unknown population initializer '%s'" % init)

    # Use LHS to initialize any "free" parameters
    # TODO: find a better way to "free" parameters on --resume/--pars
    undefined = getattr(problem, "undefined", None)
    if undefined is not None:
        del problem.undefined
        population[:, undefined] = lhs_init(pop_size, initial[undefined], bounds[:, undefined], use_point=False)

    return population


def lhs_init(n, initial, bounds, use_point=False):
    # type: (int, np.ndarray, np.ndarray, bool, float) -> np.ndarray
    """
    Latin hypercube sampling.

    Returns an array whose columns and rows each have *n* samples from
    equally spaced bins between *bounds=(xmin, xmax)* for the column.
    Unlike random, this method guarantees a certain amount of coverage
    of the parameter space.  Consider, though that the diagonal matrix
    satisfies the LHS condition, and you can see that the guarantees are
    not very strong.  A better methods, similar to sudoku puzzles, would
    guarantee coverage in each block of the matrix, but this is not
    yet implmeneted.

    If *use_point* is True, then the current value of the parameters
    is returned as the first point in the population, preserving the the
    LHS property.
    """
    xmin, xmax = bounds

    # Define the size of xmin
    nvar = len(xmin)

    # Initialize array ran with random numbers
    ran = np.random.rand(n, nvar)

    # Initialize array s with zeros
    s = empty((n, nvar))

    # Now fill s
    for j in range(nvar):
        # Indefinite and semidefinite ranges need to be constrained.  Use
        # the initial value of the parameter as a hint.
        low, high = xmin[j], xmax[j]
        if np.isinf(low) and np.isinf(high):
            if initial[j] < 0.0:
                low, high = 2.0 * initial[j], 0.0
            elif initial[j] > 0.0:
                low, high = 0.0, 2.0 * initial[j]
            else:
                low, high = -1.0, 1.0
        elif np.isinf(low):
            if initial[j] != high:
                low, high = high - 2.0 * abs(high - initial[j]), high
            else:
                low, high = high - 2.0, high
        elif np.isinf(high):
            if initial[j] != high:
                low, high = low, low + 2.0 * abs(initial[j] - low)
            else:
                low, high = low, low + 2.0
        else:
            pass  # low, high = low, high

        if use_point:
            # Put current value at position 0 in population
            s[0, j] = clip(initial[j], low, high)
            # Find which bin the current value belongs in
            xidx = int(n * (s[0, j] - low) / (high - low))
            # Generate random permutation of remaining bins
            perm = np.random.permutation(n - 1)
            perm[perm >= xidx] += 1  # exclude current value bin
            idx = slice(1, None)
        else:
            # Random permutation of bins
            perm = np.random.permutation(n)
            idx = slice(0, None)

        # Assign random value within each bin
        p = (perm + ran[idx, j]) / n
        s[idx, j] = low + p * (high - low)

    return s


def cov_init(n, initial, bounds, use_point=False, cov=None, dx=None):
    # type: (int, np.ndarray, np.ndarray, bool, Optional[np.ndarray], Optional[np.ndarray]) -> np.ndarray
    """
    Initialize *n* sets of random variables from a gaussian model.

    The center is at *x* with an uncertainty ellipse specified by the
    1-sigma independent uncertainty values *dx* or the full covariance
    matrix uncertainty *cov*.

    For example, create an initial population for 20 sequences for a
    model with local minimum x with covariance matrix C::

        pop = cov_init(cov=C, pars=p, n=20)

    If *use_point* is True, then the current value of the parameters
    is returned as the first point in the population.
    """
    # return mean + dot(RNG.randn(n,len(mean)), chol(cov))
    if cov is None:
        if dx is None:
            dx = _get_scale_factor(0.2, bounds, initial)
            # print("= dx",dx)
        cov = diag(dx**2)
    xmin, xmax = bounds
    initial = clip(initial, xmin, xmax)
    population = np.random.multivariate_normal(mean=initial, cov=cov, size=n)
    population = reflect(population, xmin, xmax)
    if use_point:
        population[0] = initial
    return population


def random_init(n, initial, bounds, use_point=False, problem=None):
    """
    Generate a random population from the problem parameters.

    Values are selected at random from the bounds of the problem according
    to the underlying probability density of each parameter.  Uniform
    semi-definite and indefinite bounds use the standard normal distribution
    for the underlying probability, with a scale factor determined by the
    initial value of the parameter.

    If *use_point* is True, then the current value of the parameters
    is returned as the first point in the population.
    """
    population = np.ascontiguousarray(problem.randomize(n))
    if use_point:
        population[0] = clip(initial, *bounds)
    return population


def eps_init(n, initial, bounds, use_point=False, eps=1e-6):
    # type: (int, np.ndarray, np.ndarray, bool, float) -> np.ndarray
    """
    Generate a random population using an epsilon ball around the current
    value.

    Since the initial population is contained in a small volume, this
    method is useful for exploring a local minimum around a point.  Over
    time the ball will expand to fill the minimum, and perhaps tunnel
    through barriers to nearby minima given enough burn-in time.

    eps is in proportion to the bounds on the parameter, or the current
    value of the parameter if the parameter is unbounded.  This gives the
    initialization a bit of scale independence.

    If *use_point* is True, then the current value of the parameters
    is returned as the first point in the population.
    """
    # Set the scale from the bounds, or from the initial value if the value
    # is unbounded.
    xmin, xmax = bounds
    scale = _get_scale_factor(eps, bounds, initial)
    # print("= scale", scale)
    initial = clip(initial, xmin, xmax)
    population = initial + scale * (2 * np.random.rand(n, len(xmin)) - 1)
    population = reflect(population, xmin, xmax)
    if use_point:
        population[0] = initial
    return population


def reflect(v, low, high):
    """
    Reflect v off the boundary, then clip to be sure it is within bounds
    """
    index = v < low
    v[index] = (2 * low - v)[index]
    index = v > high
    v[index] = (2 * high - v)[index]
    return clip(v, low, high)


def _get_scale_factor(scale, bounds, initial):
    # type: (float, np.ndarray, np.ndarray) -> np.ndarray
    xmin, xmax = bounds
    dx = (xmax - xmin) * scale  # type: np.ndarray
    dx[isinf(dx)] = abs(initial[isinf(dx)]) * scale
    dx[~isfinite(dx)] = scale
    dx[dx == 0] = scale
    # print("min,max,dx",xmin,xmax,dx)
    return dx


def demo_init(seed=1):
    # type: (Optional[int]) -> None
    from . import util
    from .bounds import init_bounds

    class Problem(object):
        def __init__(self, initial, bounds):
            self.initial = initial
            self._bounds = bounds

        def getp(self):
            return self.initial

        def bounds(self):
            return self._bounds

        def cov(self):
            return None

        def randomize(self, n=1):
            target = self.initial.copy()
            target[~isfinite(target)] = 1.0
            result = [init_bounds(pair).random(n, v) for v, pair in zip(self.initial, self._bounds.T)]
            return np.array(result).T

    bounds = np.array([(2.0, inf), (-inf, -2.0), (-inf, inf), (5.0, 6.0), (-2.0, 3.0)]).T
    # generate takes care of bad values
    # low = np.array([-inf]*5)
    # high = np.array([inf]*5)
    # bad = np.array([np.nan]*5)
    zero = np.array([0.0] * 5)
    below = np.array([-2.0, -4.0, -2.0, -3.0, -4.0])
    above = np.array([3.0, 4.0, 2.0, 8.0, 5.0])
    small = np.array([2.000001, -2.000001, 0.000001, 5.000001, -0.000001])
    large = np.array([2000001.0, -2000001.0, 2000001.0, 5.5, -2.000001])
    middle = np.array([100.0, -100.0, 100.0, 5.5, 0.5])
    starting_points = "zero below above small large middle".split()
    np.set_printoptions(linewidth=100000)
    with util.push_seed(seed):
        for init_type in ("cov", "random", "eps", "lhs"):
            print("bounds:")
            print(bounds)
            for name in starting_points:
                initial = locals()[name]
                M = Problem(initial, bounds)
                pop = generate(problem=M, init=init_type, pop=1)
                print("%s init from %s" % (init_type, name), str(initial))
                print(pop)


if __name__ == "__main__":
    demo_init(seed=None)
